from modelspec.utils import evaluate
from modelspec.utils import save_to_json_file
from modelspec.utils import locate_file
from modelspec.utils import parse_list_like

from neuromllite.utils import load_network
from neuromllite.utils import print_v
from neuromllite.utils import get_pops_vs_cell_indices_seg_ids


import numpy as np
import os
import random

import lems.api as lems  # from pylems

DEFAULT_NET_GENERATION_SEED = 1234
DEFAULT_SIMULATION_SEED = 5678



def __TEMP_CHECK_IF_SET(value):
    """
    Should just be a check for None! Update in modelspec...
    """
    if value is None: return False
    if value == '': return False
    return True

def _get_rng_for_network(nl_model):
    """
    Get a random number generator and the seed used to generate it
    """
    seed = (
        nl_model.seed
        if nl_model and nl_model.seed is not None
        else DEFAULT_NET_GENERATION_SEED
    )
    rng = random.Random(seed)

    return rng, seed


def generate_network(
    nl_model,
    handler,
    always_include_props=False,
    include_connections=True,
    include_inputs=True,
    base_dir=None,
):
    """
    Generate the network model as described in NeuroMLlite in a specific handler,
    e.g. NeuroMLHandler, PyNNHandler, etc.
    """

    pop_locations = {}
    cell_objects = {}
    synapse_objects = {}

    print_v(
        "Starting net generation for %s%s..."
        % (nl_model.id, " (base dir: %s)" % base_dir if base_dir else "")
    )

    rng, seed = _get_rng_for_network(nl_model)

    if nl_model.network_reader:

        exec(
            "from neuromllite.%s import %s"
            % (nl_model.network_reader.type, nl_model.network_reader.type)
        )

        network_reader = eval("%s()" % (nl_model.network_reader.type))
        network_reader.parameters = nl_model.network_reader.parameters

        network_reader.parse(handler)
        pop_locations = network_reader.get_locations()

    else:
        from neuromllite import __version__ as nmlv

        notes = "Generated by NeuroMLlite v%s" % nmlv
        notes += "\n    Generated network: %s" % nl_model.id
        notes += "\n    Generation seed: %i" % (seed)
        if nl_model.parameters:
            notes += "\n    NeuroMLlite parameters: "
            for p in sorted(nl_model.parameters.keys()):
                if not p == "__builtins__":
                    notes += "\n        %s = %s" % (p, nl_model.parameters[p])
        handler.handle_document_start(nl_model.id, notes)
        temperature = "%sdegC" % nl_model.temperature if nl_model.temperature is not None else None
        handler.handle_network(nl_model.id, nl_model.notes, temperature=temperature)

    nml2_doc_temp = _extract_pynn_components_to_neuroml(nl_model)

    for c in nl_model.cells:

        if c.neuroml2_source_file:
            from pyneuroml import pynml

            nml2_doc = pynml.read_neuroml2_file(
                locate_file(c.neuroml2_source_file, base_dir), include_includes=True
            )
            cell_objects[c.id] = nml2_doc.get_by_id(c.id)

        if c.pynn_cell:
            cell_objects[c.id] = nml2_doc_temp.get_by_id(c.id)

    for s in nl_model.synapses:
        if s.neuroml2_source_file:
            from pyneuroml import pynml

            nml2_doc = pynml.read_neuroml2_file(
                locate_file(s.neuroml2_source_file, base_dir), include_includes=True
            )
            synapse_objects[s.id] = nml2_doc.get_by_id(s.id)

        if hasattr(s, 'pynn_synapse') and s.pynn_synapse is not None:
            synapse_objects[s.id] = nml2_doc_temp.get_by_id(s.id)

    for p in nl_model.populations:

        size = evaluate(p.size, nl_model.parameters)
        properties = p.properties if p.properties else {}

        if p.random_layout:
            properties["region"] = p.random_layout.region
        if p.relative_layout:
            properties["region"] = p.relative_layout.region

        if (
            not p.random_layout
            and not p.single_location
            and not p.relative_layout
            and not always_include_props
        ):

            # If there are no positions (abstract network), and <property>
            # is added to <population>, jLems doesn't like it... (it has difficulty
            # interpreting pop0[0]/v, etc.)
            # So better not to give properties...
            properties = {}

        if p.notes:
            handler.handle_population(
                p.id,
                p.component,
                size,
                cell_objects[p.component] if p.component in cell_objects else None,
                properties=properties,
                notes=p.notes,
            )
        else:
            handler.handle_population(
                p.id,
                p.component,
                size,
                cell_objects[p.component] if p.component in cell_objects else None,
                properties=properties,
            )

        pop_locations[p.id] = np.zeros((size, 3))

        for i in range(size):
            if p.random_layout:
                region = nl_model.get_child(p.random_layout.region, "regions")

                x = region.x + rng.random() * region.width
                y = region.y + rng.random() * region.height
                z = region.z + rng.random() * region.depth
                pop_locations[p.id][i] = (x, y, z)

                handler.handle_location(i, p.id, p.component, x, y, z)

            if p.single_location:
                loc = p.single_location.location
                x = loc.x
                y = loc.y
                z = loc.z
                pop_locations[p.id][i] = (x, y, z)

                handler.handle_location(i, p.id, p.component, x, y, z)

            if p.relative_layout:
                print_v("Generating population with layout: %s" % p.relative_layout)
                region = nl_model.get_child(p.relative_layout.region, "regions")
                x = p.relative_layout.x + region.x
                y = p.relative_layout.y + region.y
                z = p.relative_layout.z + region.z
                pop_locations[p.id][i] = (x, y, z)
                handler.handle_location(i, p.id, p.component, x, y, z)

        if hasattr(handler, "finalise_population"):
            handler.finalise_population(p.id)

    if include_connections:
        for p in nl_model.projections:

            type = p.type if p.type else "projection"

            synapse_obj = (
                synapse_objects[p.synapse] if p.synapse in synapse_objects else None
            )
            pre_synapse_obj = (
                synapse_objects[p.pre_synapse]
                if p.pre_synapse in synapse_objects
                else None
            )

            delay = evaluate(p.delay, nl_model.parameters, rng) if p.delay else 0
            weight = evaluate(p.weight, nl_model.parameters, rng) if p.weight else 1

            if weight != 0:
                handler.handle_projection(
                    p.id,
                    p.presynaptic,
                    p.postsynaptic,
                    p.synapse,
                    synapse_obj=synapse_obj,
                    pre_synapse_obj=pre_synapse_obj,
                    type=type,
                )

                if pre_synapse_obj is None and p.pre_synapse == "silentSyn_%s" % p.id:
                    if hasattr(handler, "nml_doc"):
                        for ss in handler.nml_doc.silent_synapses:
                            if ss.id == "silentSyn_%s" % p.id:
                                handler.nml_doc.silent_synapses.remove(
                                    ss
                                )  # A default silent synapse was generated by libNeuroML

                conn_count = 0
                if p.random_connectivity:
                    for pre_i in range(len(pop_locations[p.presynaptic])):
                        for post_i in range(len(pop_locations[p.postsynaptic])):
                            flip = rng.random()
                            # print("Is cell %i conn to %i, prob %s - %s"%(pre_i, post_i, flip, p.random_connectivity.probability))
                            if flip < evaluate(
                                p.random_connectivity.probability,
                                nl_model.parameters,
                                rng,
                            ):

                                delay = (
                                    evaluate(p.delay, nl_model.parameters, rng)
                                    if p.delay
                                    else 0
                                )
                                weight = (
                                    evaluate(p.weight, nl_model.parameters, rng)
                                    if p.weight
                                    else 1
                                )
                                # print_v("Adding connection %i with weight: %s, delay: %s"%(conn_count, weight, delay))
                                handler.handle_connection(
                                    p.id,
                                    conn_count,
                                    p.presynaptic,
                                    p.postsynaptic,
                                    p.synapse,
                                    pre_i,
                                    post_i,
                                    preSegId=0,
                                    preFract=0.5,
                                    postSegId=0,
                                    postFract=0.5,
                                    delay=delay,
                                    weight=weight,
                                )
                                conn_count += 1

                if p.convergent_connectivity:

                    for post_i in range(len(pop_locations[p.postsynaptic])):

                        for count in range(int(p.convergent_connectivity.num_per_post)):
                            found = False
                            while not found:
                                pre_i = int(
                                    rng.random() * len(pop_locations[p.presynaptic])
                                )
                                if p.presynaptic == p.postsynaptic and pre_i == post_i:
                                    found = False
                                else:
                                    found = True

                            print_v(
                                "Adding connection %i (%i->%i; %i to %s of post) with weight: %s, delay: %s"
                                % (
                                    conn_count,
                                    pre_i,
                                    post_i,
                                    count,
                                    p.convergent_connectivity.num_per_post,
                                    weight,
                                    delay,
                                )
                            )
                            handler.handle_connection(
                                p.id,
                                conn_count,
                                p.presynaptic,
                                p.postsynaptic,
                                p.synapse,
                                pre_i,
                                post_i,
                                preSegId=0,
                                preFract=0.5,
                                postSegId=0,
                                postFract=0.5,
                                delay=delay,
                                weight=weight,
                            )
                            conn_count += 1

                elif p.one_to_one_connector:
                    for i in range(
                        min(
                            len(pop_locations[p.presynaptic]),
                            len(pop_locations[p.postsynaptic]),
                        )
                    ):

                        # print_v("Adding connection %i with weight: %s, delay: %s"%(conn_count, weight, delay))
                        handler.handle_connection(
                            p.id,
                            conn_count,
                            p.presynaptic,
                            p.postsynaptic,
                            p.synapse,
                            i,
                            i,
                            preSegId=0,
                            preFract=0.5,
                            postSegId=0,
                            postFract=0.5,
                            delay=delay,
                            weight=weight,
                        )
                        conn_count += 1

                handler.finalise_projection(
                    p.id, p.presynaptic, p.postsynaptic, p.synapse
                )

    if include_inputs:
        for input in nl_model.inputs:

            handler.handle_input_list(
                input.id,
                input.population,
                input.input_source,
                size=0,
                input_comp_obj=None,
            )

            input_count = 0

            if __TEMP_CHECK_IF_SET(input.cell_ids):
                if __TEMP_CHECK_IF_SET(input.percentage):
                    raise Exception(
                        "On the input: %s, only one of percentage or cell_ids is allowed"
                        % input
                    )
                for i in input.cell_ids:
                    weight = input.weight if input.weight else 1

                    if input.number_per_cell and input.segment_ids:
                        raise Exception(
                            "On input: %s, only one of number_per_cell or segment_ids is allowed"
                            % input
                        )

                    if input.number_per_cell:
                        number_per_cell = evaluate(
                            input.number_per_cell, nl_model.parameters
                        )
                        seg_ids = [0] * number_per_cell
                    elif input.segment_ids:
                        seg_ids = parse_list_like(input.segment_ids)
                    else:
                        seg_ids = [0]

                    for seg_id in seg_ids:
                        handler.handle_single_input(
                            input.id,
                            input_count,
                            i,
                            segId=seg_id,
                            weight=evaluate(weight, nl_model.parameters),
                        )
                        input_count += 1

            if __TEMP_CHECK_IF_SET(input.percentage):
                if __TEMP_CHECK_IF_SET(input.cell_ids):
                    raise Exception(
                        "On input: %s, only one of percentage or cell_ids is allowed"
                        % input
                    )
                for i in range(len(pop_locations[input.population])):
                    flip = rng.random()
                    weight = input.weight if input.weight else 1
                    if flip * 100.0 < input.percentage:

                        if input.number_per_cell and input.segment_ids:
                            raise Exception(
                                "On input: %s, only one of number_per_cell or segment_ids is allowed"
                                % input
                            )

                        if input.number_per_cell:
                            number_per_cell = evaluate(
                                input.number_per_cell, nl_model.parameters
                            )
                            seg_ids = [0] * number_per_cell
                        elif input.segment_ids:
                            seg_ids = parse_list_like(input.segment_ids)
                        else:
                            seg_ids = [0]

                        for seg_id in seg_ids:
                            handler.handle_single_input(
                                input.id,
                                input_count,
                                i,
                                segId=seg_id,
                                weight=evaluate(weight, nl_model.parameters),
                            )
                            input_count += 1

            handler.finalise_input_source(input.id)

    if hasattr(handler, "finalise_document"):
        handler.finalise_document()
    else:
        pass


def check_to_generate_or_run(argv, sim):
    """
    Useful method for calling in main method after network and simulation are
    generated, to handle some standard export options like -jnml, -graph etc.
    """

    print_v(
        "Checking arguments: %s to see whether anything should be run in simulation %s (net: %s)..."
        % (argv, sim.id, sim.network)
    )

    if len(argv) == 1:
        print_v("No arguments found. Currently supported export formats:")
        print_v(
            "   -nml |  -nmlh5 | -jnml | -jnmlnrn | -jnmlnetpyne | -netpyne | -pynnnrn "
            + "| -pynnnest | -pynnbrian | -pynnneuroml | -sonata | -matrix[1-2] | -graph[1-6 n/d/f/c]"
        )

    if "-pynnnest" in argv:
        generate_and_run(sim, simulator="PyNN_NEST")

    elif "-pynnnrn" in argv:
        generate_and_run(sim, simulator="PyNN_NEURON")

    elif "-pynnbrian" in argv:
        generate_and_run(sim, simulator="PyNN_Brian")

    elif "-arbor" in argv:
        generate_and_run(sim, simulator="Arbor")

    elif "-bindsnet" in argv:
        generate_and_run(sim, simulator="BindsNET")

    elif "-jnml" in argv:
        generate_and_run(sim, simulator="jNeuroML")

    elif "-jnmlnrn" in argv:
        generate_and_run(sim, simulator="jNeuroML_NEURON")

        """
        work in progress...
    elif '-jnmlbrian2' in argv:
        generate_and_run(sim, simulator='jNeuroML_Brian2')
"""

    elif "-netpyne" in argv:
        generate_and_run(sim, simulator="NetPyNE")

    elif "-pynnneuroml" in argv:
        generate_and_run(sim, simulator="PyNN_NeuroML")

    elif "-sonata" in argv:
        generate_and_run(sim, simulator="sonata")

    elif "-pnl" in argv or "-psyneulink" in argv:
        generate_and_run(sim, simulator="psyneulink")

    elif "-mdf" in argv:
        generate_and_run(sim, simulator="mdf")

    elif "-nml" in argv or "-neuroml" in argv:

        network = load_network(sim.network)
        generate_neuroml2_from_network(network, simulation=sim, validate=True)

    elif "-nmlh5" in argv or "-neuromlh5" in argv:

        network = load_network(sim.network)
        generate_neuroml2_from_network(
            network, simulation=sim, validate=True, format="hdf5"
        )

    else:
        for a in argv:

            if "-jnmlnetpyne" in a:
                num_processors = 1
                if len(a) > len("-jnmlnetpyne"):
                    num_processors = int(a[12:])
                generate_and_run(
                    sim, simulator="jNeuroML_NetPyNE", num_processors=num_processors
                )
            elif "graph" in a:  # e.g. -graph3c
                show_graph = True
                if "-nogui" in argv:
                    show_graph = False
                generate_and_run(sim, simulator=a[1:], nogui= not show_graph)  # Will not "run" obviously...
            elif "matrix" in a:  # e.g. -matrix2
                generate_and_run(sim, simulator=a[1:])  # Will not "run" obviously...


def _extract_pynn_components_to_neuroml(nl_model, nml_doc=None):
    """
    Parse the NeuroMLlite description for cell, synapses and inputs described as
    PyNN elements (e.g. IF_cond_alpha, DCSource) and parameters, and convert
    these to the equivalent elements in a NeuroMLDocument
    """

    import copy

    if nml_doc == None:
        from neuroml import NeuroMLDocument

        nml_doc = NeuroMLDocument(id="temp")

    for c in nl_model.cells:
        if c.pynn_cell:

            if nml_doc.get_by_id(c.id) == None:
                import pyNN.neuroml

                cell_params = copy.deepcopy(c.parameters) if c.parameters else {}
                for p in cell_params:
                    cell_params[p] = evaluate(cell_params[p], nl_model.parameters)
                for proj in nl_model.projections:

                    synapse = nl_model.get_child(proj.synapse, "synapses")
                    post_pop = nl_model.get_child(proj.postsynaptic, "populations")
                    if post_pop.component == c.id:
                        # print("--------- Cell %s in post pop %s of %s uses %s"%(c.id,post_pop.id, proj.id, synapse))

                        if synapse.pynn_receptor_type == "excitatory":
                            post = "_E"
                        elif synapse.pynn_receptor_type == "inhibitory":
                            post = "_I"
                        for p in synapse.parameters:
                            cell_params["%s%s" % (p, post)] = evaluate(
                                synapse.parameters[p], nl_model.parameters
                            )

                temp_cell = eval("pyNN.neuroml.%s(**cell_params)" % c.pynn_cell)
                if c.pynn_cell != "SpikeSourcePoisson":
                    temp_cell.default_initial_values["v"] = temp_cell.parameter_space[
                        "v_rest"
                    ].base_value

                cell_id = temp_cell.add_to_nml_doc(nml_doc, None)
                cell = nml_doc.get_by_id(cell_id)
                cell.id = c.id

    for s in nl_model.synapses:
        if nml_doc.get_by_id(s.id) == None:

            if s.pynn_synapse_type and s.pynn_receptor_type:
                import neuroml

                tau_syn = evaluate(s.parameters["tau_syn"], nl_model.parameters)
                if "e_rev" in s.parameters:
                    e_rev = evaluate(s.parameters["e_rev"], nl_model.parameters)
                if s.pynn_synapse_type == "cond_exp":
                    syn = neuroml.ExpCondSynapse(id=s.id, tau_syn=tau_syn, e_rev=e_rev)
                    nml_doc.exp_cond_synapses.append(syn)
                elif s.pynn_synapse_type == "cond_alpha":
                    syn = neuroml.AlphaCondSynapse(
                        id=s.id, tau_syn=tau_syn, e_rev=e_rev
                    )
                    nml_doc.alpha_cond_synapses.append(syn)
                elif s.pynn_synapse_type == "curr_exp":
                    syn = neuroml.ExpCurrSynapse(id=s.id, tau_syn=tau_syn)
                    nml_doc.exp_curr_synapses.append(syn)
                elif s.pynn_synapse_type == "curr_alpha":
                    syn = neuroml.AlphaCurrSynapse(id=s.id, tau_syn=tau_syn)
                    nml_doc.alpha_curr_synapses.append(syn)

    for i in nl_model.input_sources:

        if i.pynn_input:
            import pyNN.neuroml

            input_params = copy.deepcopy(i.parameters) if i.parameters else {}
            for ip in input_params:
                input_params[ip] = evaluate(input_params[ip], nl_model.parameters)
            temp_input = eval("pyNN.neuroml.%s(**input_params)" % (i.pynn_input))
            # print('%s'%(['%s->%s'%(p.id,p.amplitude) for p in nml_doc.pulse_generators]))
            pg_id = temp_input.add_to_nml_doc(nml_doc, None)
            # print('>> %s ±   %s; %s; %s'%(i,temp_input,pg_id,['%s->%s'%(p.id,p.amplitude) for p in nml_doc.pulse_generators]))
            pg = nml_doc.get_by_id(pg_id)
            pg.id = i.id
            # print('%s'%(['%s->%s'%(p.id,p.amplitude) for p in nml_doc.pulse_generators]))

    return nml_doc


def generate_neuroml2_from_network(
    nl_model,
    nml_file_name=None,
    print_summary=True,
    format="xml",
    base_dir=None,
    copy_included_elements=False,
    target_dir=None,
    validate=False,
    simulation=None,
):
    """
    Generate and save NeuroML2 file (in either XML or HDF5 format) from the
    NeuroMLlite description
    """

    print_v(
        "Generating NeuroML2 for %s%s..."
        % (
            nl_model.id,
            " (base dir: %s; target dir: %s)" % (base_dir, target_dir)
            if base_dir or target_dir
            else "",
        )
    )

    import neuroml
    from neuroml.hdf5.NetworkBuilder import NetworkBuilder

    neuroml_handler = NetworkBuilder()

    generate_network(nl_model, neuroml_handler, base_dir=base_dir)

    nml_doc = neuroml_handler.get_nml_doc()

    extra_lems_components = lems.Model()
    extra_lems_file = "%s__lems.xml" % nl_model.id

    if simulation is not None:
        if simulation.dt is not None:
            nml_doc.networks[0].properties.append(
                neuroml.Property("recommended_dt_ms", simulation.dt)
            )
        if simulation.duration is not None:
            nml_doc.networks[0].properties.append(
                neuroml.Property("recommended_duration_ms", simulation.duration)
            )

    for i in nl_model.input_sources:

        if i.lems_source_file:
            fname = locate_file(i.lems_source_file, base_dir)

            print_v(
                "Need to use for this input parameters: %s in %s from %s (%s); will place in %s"
                % (i.parameters, i.id, i.lems_source_file, fname, extra_lems_file)
            )
            model = lems.Model()
            model.import_from_file(fname)
            for comp in model.components:
                if i.id == comp.id:
                    print_v("Found a component: %s in %s" % (comp, fname))
                    if hasattr(i, 'parameters') and len(i.parameters) > 0:
                        for p in i.parameters:
                            comp.set_parameter(
                                p, evaluate(i.parameters[p], nl_model.parameters)
                            )
                    extra_lems_components.add(comp)

            for d in model.dimensions:
                print_v("Found a dimension: %s in %s" % (d, fname))
                extra_lems_components.add(d)
            for u in model.units:
                print_v("Found a unit: %s in %s" % (u, fname))
                extra_lems_components.add(u)
            for ct in model.component_types:
                print_v("Found a component type: %s in %s" % (ct, fname))
                extra_lems_components.add(ct)

            for inc in model.includes:
                print_v("Found an include: %s in %s" % (inc, fname))
                extra_lems_components.add(inc)

            incl = neuroml.IncludeType(extra_lems_file)
            if not incl in nml_doc.includes:
                nml_doc.includes.append(incl)

        if nml_doc.get_by_id(i.id) == None:
            if i.neuroml2_source_file:

                incl = neuroml.IncludeType(
                    locate_file(i.neuroml2_source_file, base_dir)
                )
                if not incl in nml_doc.includes:
                    nml_doc.includes.append(incl)

            if hasattr(i, 'neuroml2_input') and i.neuroml2_input is not None:
                input_params = i.parameters if i.parameters else {}

                # TODO make more generic...

                if i.neuroml2_input.lower() == "pulsegenerator":
                    input = neuroml.PulseGenerator(id=i.id)
                    nml_doc.pulse_generators.append(input)

                elif i.neuroml2_input.lower() == "pulsegeneratordl":
                    input = neuroml.PulseGeneratorDL(id=i.id)
                    nml_doc.pulse_generator_dls.append(input)

                elif i.neuroml2_input.lower() == "poissonfiringsynapse":
                    input = neuroml.PoissonFiringSynapse(id=i.id)
                    nml_doc.poisson_firing_synapses.append(input)

                for p in input_params:
                    exec(
                        'input.%s = "%s"'
                        % (p, evaluate(input_params[p], nl_model.parameters))
                    )

    for c in nl_model.cells:
        if c.neuroml2_source_file:

            incl = neuroml.IncludeType(locate_file(c.neuroml2_source_file, base_dir))
            found_cell = False
            for cell in nml_doc.cells:
                if cell.id == c.id:
                    nml_doc.cells.remove(
                        cell
                    )  # Better to use imported cell file; will have channels, etc.
                    nml_doc.includes.append(incl)
                    found_cell = True

            if not found_cell:
                for p in nl_model.populations:
                    if p.component == c.id:
                        pass

            if nml_doc.get_by_id(c.id) == None:  # cell (of any type) is not in nml_doc
                if not incl in nml_doc.includes:
                    nml_doc.includes.append(incl)

        if c.lems_source_file:
            fname = locate_file(c.lems_source_file, base_dir)

            print_v(
                "Need to use for this cell parameters: %s in %s from %s (%s); will place in %s"
                % (c.parameters, c.id, c.lems_source_file, fname, extra_lems_file)
            )
            model = lems.Model()
            model.import_from_file(fname)
            for comp in model.components:
                if c.id == comp.id:
                    print_v("Found component: %s in %s" % (comp, fname))
                    if c.parameters is not None and len(c.parameters) > 0:
                        for p in c.parameters:
                            comp.set_parameter(
                                p, evaluate(c.parameters[p], nl_model.parameters)
                            )
                    extra_lems_components.add(comp)

            for d in model.dimensions:
                print_v("Found a dimension: %s in %s" % (d, fname))
                extra_lems_components.add(d)
            for u in model.units:
                print_v("Found a unit: %s in %s" % (u, fname))
                extra_lems_components.add(u)
            for ct in model.component_types:
                print_v("Found component type: %s in %s" % (ct, fname))
                extra_lems_components.add(ct)
            for inc in model.includes:
                print_v("Found include: %s in %s" % (inc, fname))
                extra_lems_components.add(inc)

            incl = neuroml.IncludeType(extra_lems_file)
            if not incl in nml_doc.includes:
                nml_doc.includes.append(incl)

        if c.neuroml2_cell:

            cell_params = c.parameters if c.parameters else {}
            # TODO make more generic...
            if c.neuroml2_cell.lower() == "spikegenerator":
                cell = neuroml.SpikeGenerator(id=c.id)
                nml_doc.spike_generators.append(cell)
            elif c.neuroml2_cell.lower() == "spikegeneratorpoisson":
                cell = neuroml.SpikeGeneratorPoisson(id=c.id)
                nml_doc.spike_generator_poissons.append(cell)
            elif c.neuroml2_cell.lower() == "spikegeneratorrefpoisson":
                cell = neuroml.SpikeGeneratorRefPoisson(id=c.id)
                nml_doc.spike_generator_ref_poissons.append(cell)
            elif c.neuroml2_cell.lower() == "izhikevich2007cell":
                cell = neuroml.Izhikevich2007Cell(id=c.id)
                nml_doc.izhikevich2007_cells.append(cell)
            elif c.neuroml2_cell.lower() == "iafcell":
                cell = neuroml.IafCell(id=c.id)
                nml_doc.iaf_cells.append(cell)
            elif c.neuroml2_cell.lower() == "fitzhughnagumo1969cell":
                cell = neuroml.FitzHughNagumo1969Cell(id=c.id)
                nml_doc.fitz_hugh_nagumo1969_cells.append(cell)
            else:
                raise Exception(
                    "The neuroml2_cell: %s is not yet supported here..."
                    % c.neuroml2_cell
                )

            for p in cell_params:
                exec(
                    'cell.%s = "%s"'
                    % (p, evaluate(cell_params[p], nl_model.parameters))
                )

    for s in nl_model.synapses:
        if nml_doc.get_by_id(s.id) == None:
            if s.neuroml2_source_file:
                incl = neuroml.IncludeType(
                    locate_file(s.neuroml2_source_file, base_dir)
                )
                if not incl in nml_doc.includes:
                    nml_doc.includes.append(incl)

            if s.lems_source_file:
                fname = locate_file(s.lems_source_file, base_dir)

                print_v(
                    "Need to use for this synapse parameters: %s in %s from %s (%s); will place in %s"
                    % (s.parameters, s.id, s.lems_source_file, fname, extra_lems_file)
                )
                model = lems.Model()
                model.import_from_file(fname)
                for comp in model.components:
                    if s.id == comp.id:
                        print_v("Found component: %s in %s" % (comp, fname))
                        if s.parameters is not None and len(s.parameters) > 0:
                            for p in s.parameters:
                                comp.set_parameter(
                                    p, evaluate(s.parameters[p], nl_model.parameters)
                                )
                        extra_lems_components.add(comp)

                for d in model.dimensions:
                    print_v("Found a dimension: %s in %s" % (d, fname))
                    extra_lems_components.add(d)
                for u in model.units:
                    print_v("Found a unit: %s in %s" % (u, fname))
                    extra_lems_components.add(u)
                for ct in model.component_types:
                    print_v("Found component type: %s in %s" % (ct, fname))
                    extra_lems_components.add(ct)
                for inc in model.includes:
                    print_v("Found include: %s in %s" % (inc, fname))
                    extra_lems_components.add(inc)

                incl = neuroml.IncludeType(extra_lems_file)
                if not incl in nml_doc.includes:
                    nml_doc.includes.append(incl)

    # Look for and add the PyNN based elements to the NeuroMLDocument
    _extract_pynn_components_to_neuroml(nl_model, nml_doc)

    if print_summary:
        # Print info
        print_v(nml_doc.summary())

    # Save to file
    if target_dir == None:
        target_dir = base_dir
    if format == "xml":
        if not nml_file_name:
            nml_file_name = locate_file("%s.net.nml" % nml_doc.id, target_dir)
        from neuroml.writers import NeuroMLWriter

        NeuroMLWriter.write(nml_doc, nml_file_name)

    if format == "hdf5":
        if not nml_file_name:
            nml_file_name = locate_file("%s.net.nml.h5" % nml_doc.id, target_dir)
        from neuroml.writers import NeuroMLHdf5Writer

        NeuroMLHdf5Writer.write(nml_doc, nml_file_name)

    print_v("Written NeuroML to %s" % nml_file_name)

    if len(extra_lems_components.components) > 0:
        extra_lems_components.export_to_file(extra_lems_file)

    if validate and format == "xml":

        from pyneuroml import pynml

        success = pynml.validate_neuroml2(nml_file_name, verbose_validate=False)
        if success:
            print_v("Generated file is valid NeuroML2!")
        else:
            print_v("Generated file is NOT valid NeuroML2!")

    return nml_file_name, nml_doc


locations_mods_loaded_from = []


def _generate_neuron_files_from_neuroml(network, verbose=False, dir_for_mod_files=None):
    """
    Generate NEURON hoc/mod files from the NeuroML files which are marked as
    included in the NeuroMLlite description; also compiles the mod files
    """

    print_v(
        "-------------   Generating NEURON files from NeuroML for %s (default dir: %s)..."
        % (network.id, dir_for_mod_files)
    )
    nml_src_files = []

    from neuroml import NeuroMLDocument
    import neuroml.writers as writers

    temp_nml_doc = NeuroMLDocument(id="temp")

    dirs_for_mod_files = []
    if dir_for_mod_files != None:
        dirs_for_mod_files.append(os.path.abspath(dir_for_mod_files))

    for c in network.cells:
        if c.neuroml2_source_file:
            nml_src_files.append(c.neuroml2_source_file)
            dir_for_mod = os.path.dirname(os.path.abspath(c.neuroml2_source_file))
            if not dir_for_mod in dirs_for_mod_files:
                dirs_for_mod_files.append(dir_for_mod)

    for s in network.synapses:
        if s.neuroml2_source_file:
            nml_src_files.append(s.neuroml2_source_file)
            dir_for_mod = os.path.dirname(os.path.abspath(s.neuroml2_source_file))
            if not dir_for_mod in dirs_for_mod_files:
                dirs_for_mod_files.append(dir_for_mod)

    for i in network.input_sources:
        if i.neuroml2_source_file:
            nml_src_files.append(i.neuroml2_source_file)
            dir_for_mod = os.path.dirname(os.path.abspath(i.neuroml2_source_file))
            if not dir_for_mod in dirs_for_mod_files:
                dirs_for_mod_files.append(dir_for_mod)

    temp_nml_doc = _extract_pynn_components_to_neuroml(network)

    summary = temp_nml_doc.summary()

    if "IF_" in summary:
        import tempfile

        temp_nml_file = tempfile.NamedTemporaryFile(
            delete=False, suffix=".nml", dir=dir_for_mod_files
        )
        print_v("Writing temporary NML file to: %s, summary: " % temp_nml_file.name)
        print_v(summary)

        writers.NeuroMLWriter.write(temp_nml_doc, temp_nml_file.name)
        nml_src_files.append(temp_nml_file.name)

    for f in nml_src_files:
        from pyneuroml import pynml

        print_v("Generating/compiling hoc/mod files for: %s" % f)
        pynml.run_lems_with_jneuroml_neuron(
            f, nogui=True, only_generate_scripts=True, compile_mods=True, verbose=False
        )

    for dir_for_mod_files in dirs_for_mod_files:
        if not dir_for_mod_files in locations_mods_loaded_from:
            print_v(
                "Generated NEURON code; loading mechanisms from %s (cwd: %s; already loaded: %s)"
                % (dir_for_mod_files, os.getcwd(), locations_mods_loaded_from)
            )
            try:

                from neuron import load_mechanisms

                if os.getcwd() == dir_for_mod_files:
                    print_v(
                        "That's current dir => importing neuron module loads mods here..."
                    )
                else:
                    load_mechanisms(dir_for_mod_files)
                locations_mods_loaded_from.append(dir_for_mod_files)
            except:
                print_v("Failed to load mod file mechanisms...")
        else:
            print_v(
                "Already loaded mechanisms from %s (all loaded: %s)"
                % (dir_for_mod_files, locations_mods_loaded_from)
            )


def generate_and_run(
    simulation,
    simulator,
    network=None,
    return_results=False,
    base_dir=None,
    target_dir=None,
    num_processors=1,
    nogui=False,
):
    """
    Generates the network in the specified simulator and runs, if appropriate
    """

    if network == None:
        network = load_network(simulation.network)

    print_v(
        "Generating network %s and running in simulator: %s..."
        % (network.id, simulator)
    )

    try:
        if simulator == "NEURON":  # NOT YET WORKING...

            _generate_neuron_files_from_neuroml(network, dir_for_mod_files=target_dir)

            from neuromllite.NeuronHandler import NeuronHandler

            nrn_handler = NeuronHandler()

            for c in network.cells:
                if c.neuroml2_source_file:
                    src_dir = os.path.dirname(os.path.abspath(c.neuroml2_source_file))
                    nrn_handler.executeHoc('load_file("%s/%s.hoc")' % (src_dir, c.id))

            generate_network(network, nrn_handler, generate_network, base_dir)
            if return_results:
                raise NotImplementedError(
                    "Reloading results not supported in Neuron yet..."
                )

        elif simulator.lower() == "mdf":

            from neuromllite.MDFHandler import MDFHandler

            mdf_handler = MDFHandler(nl_network=network)

            generate_network(network, mdf_handler)

        elif simulator.lower() == "psyneulink":

            from neuromllite.PsyNeuLinkHandler import PsyNeuLinkHandler

            pnl_handler = PsyNeuLinkHandler(nl_network=network)

            generate_network(network, pnl_handler)
            from neuromllite import __version__ as nmlliteversion

            run_pnl_script = """# Generated by NeuroMLlite v{1}

import psyneulink as pnl

conv_json_filename = '{0}.bids-mdf.json'
with open(conv_json_filename, 'r') as infile:
    json = infile.read()

load_model = 'load_{0}_bids_mdf'
py_filename = load_model+'.py'
with open(py_filename, 'w') as outfi:
    outfi.write(pnl.generate_script_from_json(json))

exec('from %s import {0}'%load_model)

print('Ready to run network: {0} imported using %s'%py_filename)

dt = {2}
simtime = {3}

time_step_size=dt
num_trials=int(simtime/dt)

{0}.run(inputs={{}}, log=True, num_trials=num_trials)
print('Finished simulation run of %s ms'%simtime)

def generate_time_array(node, context='{0}', param='value'):
    return [entry.time.trial for entry in getattr(node.parameters, param).log[context]]

def generate_value_array(node, index, context='{0}', param='value'):
    return [float(entry.value[index]) for entry in getattr(node.parameters, param).log[context]]

import matplotlib.pyplot as plt
fig, axes = plt.subplots()

all_data_files = []

################################################################################
#          Note: this currently just plots/saved (almost) everything...
################################################################################

for node in {0}.nodes:
    var_num = 2 if node.function.__class__.__name__=='FitzHughNagumoIntegrator' else 1
    for i in range(var_num):
        x_values = generate_time_array(node)
        y_values = generate_value_array(node, i)

        dat_file = '{0}_%s_%i.dat'%(node.name,i)
        all_data_files.append(dat_file)
        fout = open(dat_file,'w')

        for index in range(len(x_values)):
            #                                                       1000 to convert ms to s
            fout.write('%s\\t%s\\n'%(x_values[index]*time_step_size/1000.0, y_values[index]))
        fout.close()

        times = [t*time_step_size/1000.0 for t in x_values]

        axes.plot(
            times,
            y_values,
            label=f'Value of {{i}} {{node.name}}, {{node.function.__class__.__name__}}'
        )

axes.set_xlabel('Time (s)')
axes.legend()

print('Data from simulation saved to: %s'%all_data_files)
plt.show()

#{0}.show_graph()
            """.format(
                network.id, nmlliteversion, simulation.dt, simulation.duration
            )

            run_pnl_file = open("run_pnl_%s.py" % network.id, "w")

            run_pnl_file.write(run_pnl_script)

            print_v("Done with PsyNeuLink...")

        elif simulator.lower() == "sonata":

            target_simulator = "NEST"

            from neuromllite.SonataHandler import SonataHandler

            sonata_handler = SonataHandler()

            generate_network(
                network, sonata_handler, always_include_props=True, base_dir=base_dir
            )

            import pyNN.neuroml

            for c in network.cells:
                if c.pynn_cell:

                    cell_params = {}
                    if c.parameters:
                        for p in c.parameters:
                            cell_params[p] = evaluate(
                                c.parameters[p], network.parameters
                            )

                    temp_cell = eval("pyNN.neuroml.%s(**cell_params)" % c.pynn_cell)

                    mappings = {
                        "i_offset": "I_e",
                        "tau_m": "tau_m",
                        "cm": "C_m",
                        "tau_refrac": "t_ref",
                        "v_rest": "E_L",
                        "v_thresh": "V_th",
                        "v_reset": "V_reset",
                    }
                    scales = {"cm": 1000}

                    print_v("Converting %s (%s) to Sonata format..." % (c, temp_cell))
                    comp_file = (
                        "./components/point_neuron_models_dir/%s.json" % c.id
                    )  # Oi! Hardcoding!
                    comp_file_info = {}

                    for p in temp_cell.default_parameters:
                        print(
                            "  Param %s = %s"
                            % (p, temp_cell.parameter_space[p].base_value)
                        )
                        if p in mappings:
                            sim_name = mappings[p]
                            sim_val = temp_cell.parameter_space[p].base_value * (
                                scales[p] if p in scales else 1
                            )
                            # print('    Translated %s = %s'%(sim_name, sim_val))
                            comp_file_info[sim_name] = sim_val

                    save_to_json_file(comp_file_info, comp_file, indent=2)

            sim_file_info = {}

            sim_file_info["run"] = {}
            sim_file_info["run"]["tstop"] = simulation.duration
            sim_file_info["run"]["dt"] = simulation.dt

            sim_file_info["target_simulator"] = target_simulator
            sim_file_info["manifest"] = {}
            sim_file_info["manifest"]["$OUTPUT_DIR"] = "./output"
            sim_file_info["manifest"]["$INPUT_DIR"] = "./"
            sim_file_info["output"] = {}
            sim_file_info["output"]["output_dir"] = "$OUTPUT_DIR"
            sim_file_info["output"]["log_file"] = "log.txt"
            sim_file_info["output"]["spikes_file"] = "spikes.h5"
            sim_file_info["output"]["spikes_sort_order"] = "time"

            sim_file_info["node_sets"] = {}
            sim_file_info["node_sets"]["point_nodes"] = {}
            sim_file_info["node_sets"]["point_nodes"]["model_type"] = "point_process"

            sim_file_info["reports"] = {}
            sim_file_info["reports"]["membrane_potential"] = {}
            sim_file_info["reports"]["membrane_potential"]["cells"] = "point_nodes"
            sim_file_info["reports"]["membrane_potential"]["variable_name"] = "V_m"
            sim_file_info["reports"]["membrane_potential"][
                "module"
            ] = "multimeter_report"
            sim_file_info["reports"]["membrane_potential"]["sections"] = "soma"
            sim_file_info["reports"]["membrane_potential"]["enabled"] = True

            # print sonata_handler.input_info

            temp_nml_doc = _extract_pynn_components_to_neuroml(network)

            summary = temp_nml_doc.summary()
            from pyneuroml.pynml import convert_to_units

            sim_file_info["inputs"] = {}
            for input in sonata_handler.input_info:
                sim_file_info["inputs"][input] = {}
                input_comp = sonata_handler.input_info[input][1]
                c = temp_nml_doc.get_by_id(input_comp)
                ref = "inputset_%s_%s" % (input, input_comp)
                if c.__class__.__name__ == "PulseGenerator":
                    sim_file_info["inputs"][input]["input_type"] = "current_clamp"
                    sim_file_info["inputs"][input]["module"] = "IClamp"
                    sim_file_info["inputs"][input]["amp"] = convert_to_units(
                        c.amplitude, "pA"
                    )
                    sim_file_info["inputs"][input]["delay"] = convert_to_units(
                        c.delay, "ms"
                    )
                    sim_file_info["inputs"][input]["duration"] = convert_to_units(
                        c.duration, "ms"
                    )
                    sim_file_info["inputs"][input]["node_set"] = ref
                else:
                    raise Exception("Can't yet handle input of type: %s" % c)

                sim_file_info["node_sets"][ref] = {}
                sim_file_info["node_sets"][ref]["model_type"] = "point_process"
                sim_file_info["node_sets"][ref][
                    "population"
                ] = sonata_handler.input_info[input][0]
                # sim_file_info["node_sets"][ref]['node_id'] = sonata_handler.input_info[input][2]

            save_to_json_file(sim_file_info, "simulation_config.json", indent=2)

            run_bmtk_template = """#!/bin/env python

    import sys

    def run(config_file, simulator):

        if simulator=='NEURON':
            from bmtk.simulator import bionet
            conf = bionet.Config.from_json(config_file, validate=True)
            conf.build_env()
            net = bionet.BioNetwork.from_config(conf)
            sim = bionet.BioSimulator.from_config(conf, network=net)


        elif simulator=='NEST':
            from bmtk.simulator import pointnet
            conf = pointnet.Config.from_json(config_file)
            conf.build_env()
            net = pointnet.PointNetwork.from_config(conf)
            sim = pointnet.PointSimulator.from_config(conf, net)

        sim.run()


    if __name__ == '__main__':

            run('config.json', '%s')

            """

            run_bmtk_file = open("run_bmtk.py", "w")
            run_bmtk_file.write(run_bmtk_template % (target_simulator))

            print_v("Done with Sonata...")

        elif simulator.lower().startswith("graph"):  # Will not "run" obviously...

            from neuromllite.GraphVizHandler import GraphVizHandler, engines

            try:
                if simulator[-1].isalpha():

                    engine = engines[simulator[-1]]
                    level = int(simulator[5:-1])
                else:
                    engine = "dot"
                    level = int(simulator[5:])

            except Exception as e:
                print_v("Error parsing: %s: %s" % (simulator, e))
                print_v(
                    "Graphs of the network structure can be generated at many levels of detail (1-6, required) and laid out using GraphViz engines (d - dot (default); c - circo; n - neato; f - fdp), so use: -graph3c, -graph2, -graph4f etc."
                )
                return

            handler = GraphVizHandler(level, engine=engine, nl_network=network, view_on_render = not nogui)

            generate_network(
                network, handler, always_include_props=True, base_dir=base_dir
            )

            print_v("Done with GraphViz...")

        elif simulator.lower().startswith("matrix"):  # Will not "run" obviously...

            from neuromllite.MatrixHandler import MatrixHandler

            try:
                level = int(simulator[6:])
            except:
                print_v("Error parsing: %s" % simulator)
                print_v(
                    "Matrices of the network structure can be generated at many levels of detail (1-n, required), so use: -matrix1, -matrix2, etc."
                )
                return

            handler = MatrixHandler(level, nl_network=network)

            generate_network(
                network, handler, always_include_props=True, base_dir=base_dir
            )

            print_v("Done with MatrixHandler...")

        elif simulator == "BindsNET":

            from neuromllite.BindsNETHandler import BindsNETHandler
            import bindsnet

            print("\n   ********************************************************")
            print("   *** Warning: Support for BindsNET is very preliminary!! ***")
            print("   ********************************************************\n")

            bindsnet_handler = BindsNETHandler(network)

            generate_network(
                network, bindsnet_handler, always_include_props=True, base_dir=base_dir
            )

            monitors_v = {}
            monitors_s = {}
            for pop in bindsnet_handler.pops_vs_components:
                comp = bindsnet_handler.pops_vs_components[pop]
                bn_layer = bindsnet_handler.pops_vs_bn_layers[pop]
                print("Monitoring all spikes in %s (%s) -> %s" % (pop, comp, bn_layer))
                m = bindsnet.network.monitors.Monitor(obj=bn_layer, state_vars=["s"])
                bindsnet_handler.bn_network.add_monitor(
                    monitor=m, name="%s_spikes" % pop
                )
                monitors_s[pop] = m

                if comp != "Input":
                    print("Monitoring all v in %s (%s)" % (pop, comp))
                    mv = bindsnet.network.monitors.Monitor(bn_layer, ["v"])
                    bindsnet_handler.bn_network.add_monitor(
                        monitor=mv, name="%s_voltage" % pop
                    )
                    monitors_v[pop] = mv

            traces = {}
            events = {}

            times = [
                t / 1000.0 for t in range(int(simulation.duration / simulation.dt) + 1)
            ]
            traces["t"] = times

            bindsnet_handler.bn_network.run(inputs={}, time=simulation.duration)

            for pop in monitors_v:
                vs = monitors_v[pop].get("v").numpy()
                print("Shape of voltage results for %s: %s" % (pop, vs.shape))
                v = vs
                print(v)
                traces["%s/0" % pop] = v

            for pop in monitors_s:
                ss = monitors_s[pop].get("s").numpy()
                print("Shape of spike results for %s: %s" % (pop, ss.shape))
                s = ss
                print(s)
                # traces['%s/0'%pop]= ss

            print_v("Finished BindsNET simulation")
            _print_result_info(traces, events)

            if return_results:
                return traces, events

        elif simulator == "Arbor":

            from neuromllite.ArborHandler import ArborHandler
            import arbor

            print("\n   *********************************************************")
            print("   *** Warning: Support for Arbor is very preliminary!!! ***")
            print("   *********************************************************\n")

            arbor_handler = ArborHandler(network)

            generate_network(
                network, arbor_handler, always_include_props=True, base_dir=base_dir
            )

            arbor_recipe = arbor_handler.neuroML_arbor_recipe

            trace_pop_indices_seg_ids = get_pops_vs_cell_indices_seg_ids(
                simulation.record_traces, network
            )
            spike_pop_indices = get_pops_vs_cell_indices_seg_ids(
                simulation.record_spikes, network
            )
            """
            for pop_id in arbor_handler.populations:
                arbor_pop = arbor_handler.populations[pop_id]
                if arbor_pop.label in trace_pop_indices_seg_ids:
                    if arbor_pop.can_record('v'):
                        indices = trace_pop_indices_seg_ids[pop_id].keys()
                        print_v("Recording v in %s in cells %s" % (indices, pop_id))
                        for index in indices:
                            #arbor_handler.sim.record('v',arbor_pop[index],'PP_%s_%s_%s.pkl'%(pop_id,index,'v'))
                if arbor_pop.label in spike_pop_indices:
                    if arbor_pop.can_record('spikes'):
                        indices = spike_pop_indices[pop_id].keys()
                        print_v("Recording spikes in %s in cells %s" % (indices, pop_id))
                        for index in indices:
                            #arbor_handler.sim.record('spikes',arbor_pop[index],'PP_%s_%s.pkl'%(pop_id,'spike'))
            """

            ncells = arbor_recipe.num_cells()
            print_v(
                "Starting Arbor simulation with %i cells of duration %sms (dt: %sms)"
                % (ncells, simulation.duration, simulation.dt)
            )

            # arbor_model.run(tfinal=simulation.duration)

            # (12) Create a default execution context, domain decomposition and simulation
            context = arbor.context()
            decomp = arbor.partition_load_balance(arbor_recipe, context)
            sim = arbor.simulation(arbor_recipe, decomp, context)

            # (13) Set spike generators to record
            sim.record(arbor.spike_recording.all)

            # (14) Attach a sampler to the voltage probe on cell 0.
            handles = [
                sim.sample((gid, 0), arbor.regular_schedule(simulation.dt))
                for gid in range(ncells)
            ]
            print_v("handles: %s, %s" % (handles, range(ncells)))
            # (15) Run simulation
            sim.run(simulation.duration)

            print_v("Finished Arbor simulation")
            # arbor_handler.sim.end()

            traces = {}
            events = {}

            for pop_id in trace_pop_indices_seg_ids:
                indices = trace_pop_indices_seg_ids[pop_id].keys()
                for index in indices:

                    filename = "%s.%s.%s.v.dat" % (simulation.id, pop_id, index)
                    # all_columns = []
                    gid = arbor_recipe.get_gid(pop_id, index)

                    samples, meta = sim.samples(handles[gid])[0]

                    data = samples[:, 1]
                    times = samples[:, 0]

                    print_v(
                        "Writing data for %s[%s] (gid: %s) to %s: (%s) %s"
                        % (pop_id, index, gid, filename, meta, data)
                    )

                    f = open(filename, "w")
                    for ti in range(len(times)):
                        f.write("%s\t%s\n" % (times[ti], data[ti]))

                    ref = "%s[%i]" % (pop_id, index)

                    vm_si = [d / 1000.0 for d in data]
                    traces[ref] = vm_si
                    t_si = [t / 1000.0 for t in times]
                    traces["t"] = t_si

            for pop_id in spike_pop_indices:
                indices = spike_pop_indices[pop_id]

                filename = "%s.%s.spikes" % (simulation.id, pop_id)
                all_columns = []
                print_v(
                    "Writing spike data for %s %s to %s" % (pop_id, indices, filename)
                )
                """
                arbor_pop = arbor_handler.populations[pop_id]
                data =  arbor_pop.get_data('spikes', gather=False)
                spiketrains = data.segments[0].spiketrains

                ff = open(filename, 'w')

                for spiketrain in spiketrains:
                    source_id = spiketrain.annotations['source_id']
                    source_index = spiketrain.annotations['source_index']
                    if source_index in indices:
                        #print_v("Writing spike data for cell %s[%s] (gid: %i): %i spikes "%(arbor_pop.label,source_index, source_id, len(spiketrain)), self.verbose)
                        ref = '%s/%i/???'%(arbor_pop.label,source_index)
                        events[ref] = [t.magnitude/1000. for t in spiketrain]
                        for t in spiketrain:
                            #ff.write('%s\t%i\n'%(t.magnitude/1000.,source_index))
                            ff.write('%i\t%s\n'%(source_index,t.magnitude/1000.))
                ff.close()"""

            if return_results:
                _print_result_info(traces, events)
                return traces, events

        elif simulator.startswith("PyNN"):

            # _generate_neuron_files_from_neuroml(network)
            simulator_name = simulator.split("_")[1].lower()

            from neuromllite.PyNNHandler import PyNNHandler

            pynn_handler = PyNNHandler(simulator_name, simulation.dt, network.id)

            syn_cell_params = {}
            for proj in network.projections:

                synapse = network.get_child(proj.synapse, "synapses")
                post_pop = network.get_child(proj.postsynaptic, "populations")

                if not post_pop.component in syn_cell_params:
                    syn_cell_params[post_pop.component] = {}
                for p in synapse.parameters:
                    post = ""
                    if synapse.pynn_receptor_type == "excitatory":
                        post = "_E"
                    elif synapse.pynn_receptor_type == "inhibitory":
                        post = "_I"
                    syn_cell_params[post_pop.component]["%s%s" % (p, post)] = evaluate(
                        synapse.parameters[p], network.parameters
                    )

            cells = {}
            for c in network.cells:
                if c.pynn_cell:
                    cell_params = {}
                    if c.parameters:
                        for p in c.parameters:
                            cell_params[p] = evaluate(
                                c.parameters[p], network.parameters
                            )

                    dont_set_here = ["tau_syn_E", "e_rev_E", "tau_syn_I", "e_rev_I"]
                    for d in dont_set_here:
                        if d in c.parameters:
                            print("Problem with %s, %s" % (c.id, c.parameters))
                            raise Exception(
                                "Synaptic parameters like %s should be set " % d
                                + "in individual synapses, not in the list of parameters associated with the cell"
                            )
                    if c.id in syn_cell_params:
                        cell_params.update(syn_cell_params[c.id])
                    print_v("Creating cell with params: %s" % cell_params)
                    exec(
                        'cells["%s"] = pynn_handler.sim.%s(**cell_params)'
                        % (c.id, c.pynn_cell)
                    )

                    if c.pynn_cell != "SpikeSourcePoisson":
                        exec(
                            "cells['%s'].default_initial_values['v'] = cells['%s'].parameter_space['v_rest'].base_value"
                            % (c.id, c.id)
                        )

            pynn_handler.set_cells(cells)

            receptor_types = {}
            for s in network.synapses:
                if s.pynn_receptor_type:
                    receptor_types[s.id] = s.pynn_receptor_type

            pynn_handler.set_receptor_types(receptor_types)

            for input_source in network.input_sources:

                # TODO: change this when noisyCurrentSource is integrated into the core of NeuroML
                if input_source.pynn_input or input_source.lems_source_file:
                    pynn_handler.add_input_source(input_source, network)

            generate_network(
                network, pynn_handler, always_include_props=True, base_dir=base_dir
            )

            trace_pop_indices_seg_ids = get_pops_vs_cell_indices_seg_ids(
                simulation.record_traces, network
            )
            spike_pop_indices = get_pops_vs_cell_indices_seg_ids(
                simulation.record_spikes, network
            )

            for pop_id in pynn_handler.populations:
                pynn_pop = pynn_handler.populations[pop_id]
                if pynn_pop.label in trace_pop_indices_seg_ids:
                    if pynn_pop.can_record("v"):
                        indices = trace_pop_indices_seg_ids[pop_id].keys()
                        print_v("Recording v in %s in cells %s" % (indices, pop_id))
                        for index in indices:
                            pynn_handler.sim.record(
                                "v",
                                pynn_pop[index],
                                "PP_%s_%s_%s.pkl" % (pop_id, index, "v"),
                            )
                if pynn_pop.label in spike_pop_indices:
                    if pynn_pop.can_record("spikes"):
                        indices = spike_pop_indices[pop_id].keys()
                        print_v(
                            "Recording spikes in %s in cells %s" % (indices, pop_id)
                        )
                        for index in indices:
                            pynn_handler.sim.record(
                                "spikes",
                                pynn_pop[index],
                                "PP_%s_%s.pkl" % (pop_id, "spike"),
                            )

            print_v(
                "Starting PyNN simulation of duration %sms (dt: %sms)"
                % (simulation.duration, simulation.dt)
            )
            pynn_handler.sim.run(simulation.duration)
            print_v("Finished PyNN simulation")
            pynn_handler.sim.end()

            traces = {}
            events = {}

            if not "NeuroML" in simulator:

                for pop_id in trace_pop_indices_seg_ids:
                    pynn_pop = pynn_handler.populations[pop_id]
                    indices = trace_pop_indices_seg_ids[pop_id].keys()

                    filename = "%s.%s.v.dat" % (simulation.id, pop_id)
                    all_columns = []

                    print_v(
                        "Writing data for %s %s to %s" % (pop_id, indices, filename)
                    )

                    data = pynn_pop.get_data("v", gather=False)
                    analogsignal = data.segments[0].analogsignals[0]

                    source_ids = analogsignal.annotations["source_ids"]
                    analogsignal_index = {}
                    for sii, glob_id in enumerate(source_ids):
                        index_in_pop = pynn_pop.id_to_index(glob_id)
                        # print('Index %i in analogsignal is pynn global id %i, pop index %i'%(sii,glob_id,index_in_pop))
                        analogsignal_index[index_in_pop] = sii

                    for index in indices:
                        ref = "%s[%i]" % (pop_id, index)
                        ref = "%s/%i/???/v" % (pop_id, index)
                        traces[ref] = []
                        vm = analogsignal.transpose()[analogsignal_index[index]]

                        if len(all_columns) == 0:
                            tt = np.array(
                                [t * simulation.dt / 1000.0 for t in range(len(vm))]
                            )
                            all_columns.append(tt)
                        vm_si = np.array(vm / 1000.0)

                        traces[ref] = vm_si
                        if not "t" in traces:
                            traces["t"] = tt

                        all_columns.append(vm_si)

                        times_vm = np.array(all_columns).transpose()

                    np.savetxt(filename, times_vm, delimiter="\t", fmt="%s")

                for pop_id in spike_pop_indices:
                    pynn_pop = pynn_handler.populations[pop_id]
                    indices = spike_pop_indices[pop_id]

                    filename = "%s.%s.spikes" % (simulation.id, pop_id)
                    all_columns = []
                    print_v(
                        "Writing spike data for %s %s to %s"
                        % (pop_id, indices, filename)
                    )

                    data = pynn_pop.get_data("spikes", gather=False)
                    spiketrains = data.segments[0].spiketrains

                    ff = open(filename, "w")

                    for spiketrain in spiketrains:
                        source_id = spiketrain.annotations["source_id"]
                        source_index = spiketrain.annotations["source_index"]
                        if source_index in indices:
                            # print_v("Writing spike data for cell %s[%s] (gid: %i): %i spikes "%(pynn_pop.label,source_index, source_id, len(spiketrain)), self.verbose)
                            ref = "%s/%i/???" % (pynn_pop.label, source_index)
                            events[ref] = [t.magnitude / 1000.0 for t in spiketrain]
                            for t in spiketrain:
                                # ff.write('%s\t%i\n'%(t.magnitude/1000.,source_index))
                                ff.write(
                                    "%i\t%s\n" % (source_index, t.magnitude / 1000.0)
                                )
                    ff.close()

            if return_results:
                _print_result_info(traces, events)
                return traces, events

        elif simulator == "NetPyNE":

            if target_dir == None:
                target_dir = "./"

            _generate_neuron_files_from_neuroml(network, dir_for_mod_files=target_dir)

            from netpyne import specs
            from netpyne import sim

            # Note NetPyNE from this branch is required: https://github.com/Neurosim-lab/netpyne/tree/neuroml_updates
            from netpyne.conversion.neuromlFormat import NetPyNEBuilder

            import pprint

            pp = pprint.PrettyPrinter(depth=6)

            netParams = specs.NetParams()
            simConfig = specs.SimConfig()
            netpyne_handler = NetPyNEBuilder(
                netParams, simConfig=simConfig, verbose=True
            )

            generate_network(network, netpyne_handler, base_dir=base_dir)

            netpyne_handler.finalise()

            simConfig = specs.SimConfig()
            simConfig.tstop = simulation.duration
            simConfig.duration = simulation.duration
            simConfig.dt = simulation.dt
            simConfig.seed = simulation.seed
            simConfig.recordStep = simulation.dt

            simConfig.recordCells = ["all"]
            simConfig.recordTraces = {}

            trace_pop_indices_seg_ids = get_pops_vs_cell_indices_seg_ids(
                simulation.record_traces, network
            )

            for pop_id in trace_pop_indices_seg_ids:
                for index in trace_pop_indices_seg_ids[pop_id]:
                    seg_ids = trace_pop_indices_seg_ids[pop_id][index]
                    if seg_ids == None:
                        simConfig.recordTraces[
                            "%s.%s.%s.v" % (simulation.id, pop_id, index)
                        ] = {
                            "sec": "soma",
                            "loc": 0.5,
                            "var": "v",
                            "conds": {"pop": pop_id, "cellLabel": index},
                        }
                    else:
                        for seg_id in seg_ids:
                            simConfig.recordTraces[
                                "%s.%s.%s.%s.v" % (simulation.id, pop_id, index, seg_id)
                            ] = {
                                "sec": seg_id,
                                "loc": 0.5,
                                "var": "v",
                                "conds": {"pop": pop_id, "cellLabel": index},
                            }

            # simConfig.saveDat = True

            print_v("NetPyNE netParams: ")
            pp.pprint(netParams.todict())
            # print_v("NetPyNE simConfig: ")
            # pp.pprint(simConfig.todict())

            sim.initialize(
                netParams, simConfig
            )  # create network object and set cfg and net params

            sim.net.createPops()
            cells = (
                sim.net.createCells()
            )  # instantiate network cells based on defined populations

            for proj_id in netpyne_handler.projection_infos.keys():
                (
                    projName,
                    prePop,
                    postPop,
                    synapse,
                    ptype,
                ) = netpyne_handler.projection_infos[proj_id]
                print_v(
                    "Creating connections for %s (%s): %s->%s via %s"
                    % (projName, ptype, prePop, postPop, synapse)
                )

                preComp = netpyne_handler.pop_ids_vs_components[prePop]

                for conn in netpyne_handler.connections[projName]:

                    (
                        pre_id,
                        pre_seg,
                        pre_fract,
                        post_id,
                        post_seg,
                        post_fract,
                        delay,
                        weight,
                    ) = conn

                    # connParam = {'delay':delay,'weight':weight,'synsPerConn':1, 'sec':post_seg, 'loc':post_fract, 'threshold':threshold}
                    connParam = {
                        "delay": delay,
                        "weight": weight,
                        "synsPerConn": 1,
                        "sec": post_seg,
                        "loc": post_fract,
                    }

                    if ptype == "electricalProjection":

                        if weight != 1:
                            raise Exception(
                                "Cannot yet support inputs where weight !=1!"
                            )
                        connParam = {
                            "synsPerConn": 1,
                            "sec": post_seg,
                            "loc": post_fract,
                            "gapJunction": True,
                            "weight": weight,
                        }
                    else:
                        connParam = {
                            "delay": delay,
                            "weight": weight,
                            "synsPerConn": 1,
                            "sec": post_seg,
                            "loc": post_fract,
                        }
                        #'threshold': threshold}

                    connParam["synMech"] = synapse

                    if (
                        post_id in sim.net.gid2lid
                    ):  # check if postsyn is in this node's list of gids
                        sim.net._addCellConn(connParam, pre_id, post_id)

            stims = (
                sim.net.addStims()
            )  # add external stimulation to cells (IClamps etc)
            simData = (
                sim.setupRecording()
            )  # setup variables to record for each cell (spikes, V traces, etc)
            sim.runSim()  # run parallel Neuron simulation
            sim.gatherData()  # gather spiking data and cell info from each node
            sim.saveData()  # save params, cell info and sim output to file (pickle,mat,txt,etc)

            print_v("Finished running simulation in NetPyNE")

            print_v("Extracting results from %s" % (sim.allSimData.keys()))

            times = [t / 1000.0 for t in sim.allSimData["t"]]
            traces = {}
            traces["t"] = times
            events = {}

            for pop_id in trace_pop_indices_seg_ids:
                indices_segids = trace_pop_indices_seg_ids[pop_id]
                print("Saving in population %s: %s" % (pop_id, indices_segids))
                save_ref = "%s.%s.v.dat" % (simulation.id, pop_id)
                pop_v_file = open(save_ref, "w")
                all_v = []
                for index in indices_segids:
                    seg_ids = indices_segids[index]
                    if seg_ids == None:
                        ref = "%s.%s.%s.v" % (simulation.id, pop_id, index)
                        # print('Ref: %s, existing: %s'%(ref, sim.allSimData.keys()))
                        v = [
                            v / 1000.0
                            for v in sim.allSimData[ref][list(sim.allSimData[ref])[0]]
                        ]
                        # print _print_info(v)

                        ref = "%s/%i/???/v" % (pop_id, index)
                        traces[ref] = v
                        all_v.append(v)
                    else:
                        for seg_id in seg_ids:
                            ref = "%s.%s.%s.%s.v" % (
                                simulation.id,
                                pop_id,
                                index,
                                seg_id,
                            )
                            print(
                                "Ref: %s, existing: %s" % (ref, sim.allSimData.keys())
                            )
                            v = [
                                v / 1000.0
                                for v in sim.allSimData[ref][
                                    list(sim.allSimData[ref])[0]
                                ]
                            ]
                            # print _print_info(v)

                            ref = "%s/%i/%i/v" % (pop_id, index, seg_id)
                            traces[ref] = v
                            all_v.append(v)

                for ti in range(len(times)):
                    vs = ""
                    for v in all_v:
                        vs += "\t%s" % (v[ti])
                    pop_v_file.write("%s\t" % (times[ti]) + vs + "\n")
                pop_v_file.close()

            spike_pop_indices = sorted(
                get_pops_vs_cell_indices_seg_ids(
                    simulation.record_spikes, network
                ).keys()
            )

            if (
                return_results or True
            ):  ###########################################################
                return traces, events

        elif (
            simulator == "jNeuroML"
            or simulator == "jNeuroML_norun"
            or simulator == "jNeuroML_NEURON"
            or simulator == "jNeuroML_NetPyNE"
            or simulator == "jNeuroML_Brian2"
        ):

            from pyneuroml.lems import generate_lems_file_for_neuroml
            from pyneuroml import pynml

            lems_file_name = "LEMS_%s.xml" % simulation.id

            nml_file_name, nml_doc = generate_neuroml2_from_network(
                network, simulation=simulation, base_dir=base_dir, target_dir=target_dir
            )
            included_files = ["PyNN.xml"]
            """ Needed?
            for c in network.cells:
                if c.lems_source_file:
                    included_files.append(c.lems_source_file)

            if network.cells:
                for c in network.cells:
                    included_files.append(c.neuroml2_source_file)

            if network.synapses:
                for s in network.synapses:
                    if s.lems_source_file:
                        fname = locate_file(s.lems_source_file, base_dir)
                        # more ..?
            """
            print_v("Generating LEMS file prior to running in %s" % simulator)

            pops_plot_save = []
            pops_spike_save = []
            gen_plots_for_quantities = {}
            gen_saves_for_quantities = {}
            gen_spike_saves_for_cells = {}

            trace_pop_indices_seg_ids = get_pops_vs_cell_indices_seg_ids(
                simulation.record_traces, network
            )
            spike_pop_indices = get_pops_vs_cell_indices_seg_ids(
                simulation.record_spikes, network
            )
            rate_pop_indices = get_pops_vs_cell_indices_seg_ids(
                simulation.record_rates, network
            )

            pops = network.populations
            pops.sort(key=id)
            for p in pops:

                if p.id in spike_pop_indices:
                    pops_spike_save.append(p.id)

                    save_ref = "%s.%s.spikes" % (simulation.id, p.id)
                    gen_spike_saves_for_cells[save_ref] = []

                    for i in spike_pop_indices[p.id].keys():
                        quantity = "%s/%i/%s" % (p.id, i, p.component)
                        if not p.has_positions():
                            quantity = "%s[%i]" % (p.id, i)
                        gen_spike_saves_for_cells[save_ref].append(quantity)

                if p.id in trace_pop_indices_seg_ids:
                    plot_ref = "%s_v" % (p.id)
                    save_ref = "%s.%s.v.dat" % (simulation.id, p.id)
                    gen_plots_for_quantities[plot_ref] = []
                    gen_saves_for_quantities[save_ref] = []

                    for cell_index in trace_pop_indices_seg_ids[p.id]:
                        seg_ids = trace_pop_indices_seg_ids[p.id][cell_index]
                        print_v(
                            "Recording in pop %s, cell index %i, seg ids: %s"
                            % (p.id, cell_index, seg_ids)
                        )
                        if seg_ids == None:
                            quantity = "%s/%i/%s/v" % (p.id, cell_index, p.component)
                            if not p.has_positions():
                                quantity = "%s[%i]/v" % (p.id, cell_index)
                            gen_plots_for_quantities[plot_ref].append(quantity)
                            gen_saves_for_quantities[save_ref].append(quantity)
                        else:
                            for seg_id in seg_ids:
                                quantity = "%s/%i/%s/%s/v" % (
                                    p.id,
                                    cell_index,
                                    p.component,
                                    seg_id,
                                )
                                if not p.has_positions():
                                    quantity = "%s[%i]/%s/v" % (
                                        p.id,
                                        cell_index,
                                        seg_id,
                                    )
                                gen_plots_for_quantities[plot_ref].append(quantity)
                                gen_saves_for_quantities[save_ref].append(quantity)

                if p.id in rate_pop_indices:
                    plot_ref = "%s_r" % (p.id)
                    save_ref = "%s.%s.r.dat" % (simulation.id, p.id)
                    gen_plots_for_quantities[plot_ref] = []
                    gen_saves_for_quantities[save_ref] = []

                    for i in rate_pop_indices[p.id].keys():
                        quantity = "%s/%i/%s/r" % (p.id, i, p.component)
                        gen_plots_for_quantities[plot_ref].append(quantity)
                        gen_saves_for_quantities[save_ref].append(quantity)

                if simulation.record_variables:
                    for var in simulation.record_variables:
                        to_rec = simulation.record_variables[var]
                        if "all" in to_rec or p.id in to_rec:
                            size = evaluate(p.size, network.parameters)
                            for i in range(size):
                                quantity = "%s/%i/%s/%s" % (p.id, i, p.component, var)
                                if not p.has_positions():
                                    quantity = "%s[%i]/%s" % (p.id, i, var)
                                gen_plots_for_quantities[
                                    "%s_%i_%s" % (p.id, i, var)
                                ] = [quantity]
                                safe_var = var.replace("/", "_")
                                gen_saves_for_quantities[
                                    "%s_%i.%s.dat" % (p.id, i, safe_var)
                                ] = [quantity]

            generate_lems_file_for_neuroml(
                simulation.id,
                nml_file_name,
                network.id,
                simulation.duration,
                simulation.dt,
                lems_file_name,
                target_dir=target_dir if target_dir else ".",
                nml_doc=nml_doc,  # Use this if the nml doc has already been loaded (to avoid delay in reload)
                include_extra_files=included_files,
                gen_plots_for_all_v=False,
                plot_all_segments=False,
                gen_plots_for_quantities=gen_plots_for_quantities,  # Dict with displays vs lists of quantity paths
                gen_plots_for_only_populations=pops_plot_save,  # List of populations, all pops if = []
                gen_saves_for_all_v=False,
                save_all_segments=False,
                gen_saves_for_only_populations=pops_plot_save,  # List of populations, all pops if = []
                gen_saves_for_quantities=gen_saves_for_quantities,  # Dict with file names vs lists of quantity paths
                gen_spike_saves_for_all_somas=False,
                gen_spike_saves_for_only_populations=[],  # List of populations, all pops if = []
                gen_spike_saves_for_cells=gen_spike_saves_for_cells,  # Dict with file names vs lists of quantity paths
                spike_time_format="ID_TIME",
                copy_neuroml=True,
                lems_file_generate_seed=12345,
                report_file_name="report.%s.txt" % simulation.id,
                simulation_seed=simulation.seed
                if simulation.seed
                else DEFAULT_SIMULATION_SEED,
                verbose=True,
            )

            lems_file_name = locate_file(lems_file_name, target_dir)

            if simulator == "jNeuroML_norun":
                return lems_file_name

            if simulator == "jNeuroML":
                results = pynml.run_lems_with_jneuroml(
                    lems_file_name,
                    nogui=True,
                    load_saved_data=return_results,
                    reload_events=return_results,
                )
            elif simulator == "jNeuroML_NEURON":
                results = pynml.run_lems_with_jneuroml_neuron(
                    lems_file_name,
                    nogui=True,
                    load_saved_data=return_results,
                    reload_events=return_results,
                )
            elif simulator == "jNeuroML_NetPyNE":
                results = pynml.run_lems_with_jneuroml_netpyne(
                    lems_file_name,
                    nogui=True,
                    verbose=True,
                    load_saved_data=return_results,
                    reload_events=return_results,
                    num_processors=num_processors,
                )

            elif simulator == "jNeuroML_Brian2":
                """Work in progress..."""
                results = pynml.run_lems_with_jneuroml_brian2(
                    lems_file_name,
                    nogui=True,
                    load_saved_data=return_results,
                    reload_events=return_results,
                )

            from pyneuroml import __version__, JNEUROML_VERSION

            print_v(
                "Finished running LEMS file %s in %s (pyNeuroML v%s containing jNeuroML v%s; returning results: %s)"
                % (
                    lems_file_name,
                    simulator,
                    __version__,
                    JNEUROML_VERSION,
                    return_results,
                )
            )

            if return_results:
                traces, events = results
                _print_result_info(traces, events)
                return results  # traces, events =
    except Exception as e:
        raise e
        raise Exception(
            "Exception while trying to run in simulator %s:\n%s" % (simulator, e)
        ) from e


def _print_info(l):
    print(
        "List %s of %i elements of type %s: %s -> %s (min: %s, max: %s)"
        % (type(l), len(l), type(l[0]), l[0], l[-1], min(l), max(l))
    )


def _print_result_info(traces, events):
    """
    Print a summary of the returned (voltage) traces and spike times
    """
    count = 0
    max_print = 4
    verbose = False
    print_v("Returning %i traces:" % len(traces))
    for r in sorted(traces.keys()):
        x = traces[r]
        if not verbose and count >= max_print:
            if count == max_print:
                print_v("  ...")
                count += 1
        else:
            print_v(
                "  %s (%s): %s -> %s (min: %s, max: %s, len: %i)"
                % (r, type(x), x[0], x[-1], min(x), max(x), len(x))
            )
            count += 1

    count = 0
    print_v("Returning %i events:" % len(events))
    for r in sorted(events.keys()):
        x = events[r]
        if not verbose and count >= max_print:
            if count == max_print:
                print_v("  ...")
                count += 1
        else:
            print_v(
                "  %s: %s -> %s (len: %i)"
                % (r, x[0] if len(x) > 0 else "-", x[-1] if len(x) > 0 else "-", len(x))
            )
            count += 1



from neuromllite import Network, Cell, InputSource, Population, Synapse
from neuromllite import Projection, RandomConnectivity, Input, Simulation
import sys

################################################################################
###   Build new network

net = Network(id="Example10_Lorenz")
net.notes = "Example 10: Lorenz"
net.parameters = {"N": 1}

cell = Cell(
    id="lorenzCell", lems_source_file="test_files/Lorenz1963.xml", parameters={}
)
net.cells.append(cell)

params = {"sigma": 10, "b": 2.67, "r": 28, "x0": 1.0, "y0": 1.0, "z0": 1.0}

for p in params:
    cell.parameters[p] = p
    net.parameters[p] = params[p]


pop = Population(
    id="lorenzPop", size="1", component=cell.id, properties={"color": ".7 0 0"}
)
net.populations.append(pop)


print(net)
print(net.to_json())
new_file = net.to_json_file("%s.json" % net.id)


################################################################################
###   Build Simulation object & save as JSON

sim = Simulation(
    id="SimExample10",
    network=new_file,
    duration="1e4",
    dt="1",
    record_variables={"x": {"all": "*"}, "y": {"all": "*"}, "z": {"all": "*"}},
    plots2D={
        "X-Y": {"x_axis": "lorenzPop[0]/x", "y_axis": "lorenzPop[0]/y"},
        "Y-Z": {"x_axis": "lorenzPop[0]/y", "y_axis": "lorenzPop[0]/z"},
        "X-Z": {"x_axis": "lorenzPop[0]/x", "y_axis": "lorenzPop[0]/z"},
    },
    plots3D={
        "X-Y-Z": {
            "x_axis": "lorenzPop[0]/x",
            "y_axis": "lorenzPop[0]/y",
            "z_axis": "lorenzPop[0]/z",
        }
    },
)

sim.to_json_file()


################################################################################
###   Run in some simulators

from neuromllite.NetworkGenerator import check_to_generate_or_run
import sys

check_to_generate_or_run(sys.argv, sim)



from neuromllite import Network, Population, Projection, RandomConnectivity


################################################################################
###   Build a new network

net = Network(id="net0")
net.notes = (
    "A simple network with 2 populations & projection between them. "
    + "No info yet on what the cells are so network can't be simulated."
)

print(net)


################################################################################
###   Add some populations

p0 = Population(id="pop0", size=5, component="iaf", properties={"color": "0 .8 0"})
p1 = Population(id="pop1", size=10, component="iaf", properties={"color": "0 0 .8"})

print(p1.to_json())

net.populations.append(p0)
net.populations.append(p1)


################################################################################
###   Add a projection

net.projections.append(
    Projection(id="proj0", presynaptic=p0.id, postsynaptic=p1.id, synapse="ampa")
)

net.projections[0].random_connectivity = RandomConnectivity(probability=0.5)


################################################################################
###   Save to JSON format

print(net)
net.id = "TestNetwork"

print(net.to_json())
new_file = net.to_json_file("Example1_%s.json" % net.id)


################################################################################
###   Export to some formats, e.g. try:
###        python Example1.py -graph2

from neuromllite.NetworkGenerator import check_to_generate_or_run
from neuromllite import Simulation
import sys

check_to_generate_or_run(sys.argv, Simulation(id="SimExample1", network=new_file))



#
#
#   A class to handle events by Arbor (work in progress)...
#
#

from neuromllite.utils import print_v
from neuromllite.utils import evaluate
from neuromllite.DefaultNetworkHandler import DefaultNetworkHandler

from pyneuroml.pynml import convert_to_units

import numpy as np

import arbor


class ArborHandler(DefaultNetworkHandler):

    pops_vs_components = {}
    proj_weights = {}
    proj_delays = {}
    input_info = {}
    input_lists = {}

    """
    populations = {}
    projections = {}
    input_sources = {}

    inputs = []
    cells = {}"""

    pop_indices_vs_gids = {}

    def __init__(self, nl_network):
        print_v("Initiating Arbor...")
        self.nl_network = nl_network
        self.curr_gid = 0

    def handle_document_start(self, id, notes):
        print_v("Document: %s" % id)

    def handle_network(self, network_id, notes, temperature=None):

        print_v("Network: %s" % network_id)
        if temperature:
            print_v("  Temperature: " + temperature)
        if notes:
            print_v("  Notes: " + notes)

    def handle_population(
        self, population_id, component, size=-1, component_obj=None, properties={}
    ):

        sizeInfo = " as yet unspecified size"
        if size >= 0:
            sizeInfo = ", size: " + str(size) + " cells"
        if component_obj:
            compInfo = " (%s)" % component_obj.__class__.__name__
        else:
            compInfo = ""

        print_v(
            "Population: "
            + population_id
            + ", component: "
            + component
            + compInfo
            + sizeInfo
        )

        self.pops_vs_components[population_id] = component
        self.pop_indices_vs_gids[population_id] = {}

        """
        self.model = arbor.single_cell_model(self.cells[component])
        self.model.probe('voltage', '"center"', frequency=10000)

        exec('self.POP_%s = self.sim.Population(%s, self.cells["%s"], label="%s")'%(population_id,size,component,population_id))
        #exec('print_v(self.POP_%s)'%(population_id))
        exec('self.populations["%s"] = self.POP_%s'%(population_id,population_id))"""

    #
    #  Should be overridden to create specific cell instance
    #
    def handle_location(self, id, population_id, component, x, y, z):
        # self.printLocationInformation(id, population_id, component, x, y, z)

        self.pop_indices_vs_gids[population_id][id] = self.curr_gid
        self.curr_gid += 1

        """
        exec('self.POP_%s.positions[0][%s] = %s'%(population_id,id,x))
        exec('self.POP_%s.positions[1][%s] = %s'%(population_id,id,y))
        exec('self.POP_%s.positions[2][%s] = %s'%(population_id,id,z))"""

    def handle_projection(
        self,
        projName,
        prePop,
        postPop,
        synapse,
        hasWeights=False,
        hasDelays=False,
        type="projection",
        synapse_obj=None,
        pre_synapse_obj=None,
    ):

        synInfo = ""
        if synapse_obj:
            synInfo += " (syn: %s)" % synapse_obj.__class__.__name__

        if pre_synapse_obj:
            synInfo += " (pre comp: %s)" % pre_synapse_obj.__class__.__name__

        print_v(
            "Projection: "
            + projName
            + " ("
            + type
            + ") from "
            + prePop
            + " to "
            + postPop
            + " with syn: "
            + synapse
            + synInfo
        )

        self.proj_weights[projName] = np.zeros(
            (
                len(self.pop_indices_vs_gids[prePop]),
                len(self.pop_indices_vs_gids[postPop]),
            )
        )
        self.proj_delays[projName] = np.zeros(
            (
                len(self.pop_indices_vs_gids[prePop]),
                len(self.pop_indices_vs_gids[postPop]),
            )
        )

        """
        exec('self.projection__%s_conns = []'%(projName))"""

    #
    #  Should be overridden to handle network connection
    #
    def handle_connection(
        self,
        projName,
        id,
        prePop,
        postPop,
        synapseType,
        preCellId,
        postCellId,
        preSegId=0,
        preFract=0.5,
        postSegId=0,
        postFract=0.5,
        delay=0,
        weight=1,
    ):

        self.print_connection_information(
            projName, id, prePop, postPop, synapseType, preCellId, postCellId, weight
        )
        print_v(
            "Src cell: %d, seg: %f, fract: %f -> Tgt cell %d, seg: %f, fract: %f; weight %s, delay: %s ms"
            % (
                preCellId,
                preSegId,
                preFract,
                postCellId,
                postSegId,
                postFract,
                weight,
                delay,
            )
        )

        self.proj_weights[projName][preCellId][postCellId] = weight
        self.proj_delays[projName][preCellId][postCellId] = delay

    #
    #  Should be overridden to handle end of network connection
    #
    def finalise_projection(
        self, projName, prePop, postPop, synapse=None, type="projection"
    ):

        print_v(
            "Projection finalising: "
            + projName
            + " from "
            + prePop
            + " to "
            + postPop
            + " completed"
        )

    #
    #  Should be overridden to create input source array
    #
    def handle_input_list(
        self, inputListId, population_id, component, size, input_comp_obj=None
    ):

        self.print_input_information(inputListId, population_id, component, size)
        if input_comp_obj:
            print("Input comp: %s" % input_comp_obj)

        if size < 0:
            self.log.error(
                "Error! Need a size attribute in sites element to create spike source!"
            )
            return

        self.input_info[inputListId] = (population_id, component, input_comp_obj)
        self.input_lists[inputListId] = []

    #
    #  Should be overridden to to connect each input to the target cell
    #
    def handle_single_input(
        self, inputListId, id, cellId, segId=0, fract=0.5, weight=1
    ):

        population_id, component, input_comp_obj = self.input_info[inputListId]

        print_v(
            "Input: %s[%s] (%s), pop: %s, cellId: %i, seg: %i, fract: %f, weight: %f"
            % (inputListId, id, component, population_id, cellId, segId, fract, weight)
        )
        self.input_lists[inputListId].append((cellId, segId, fract, weight))

    #
    #  Should be overridden to to connect each input to the target cell
    #
    def finalise_input_source(self, inputName):
        print_v("Input: %s completed" % inputName)

    def finalise_document(self):
        print_v("Building recipe with: %s" % self.pop_indices_vs_gids)
        print_v("Weights: %s" % self.proj_weights)
        print_v("Delays: %s" % self.proj_delays)
        print_v("Inputs: %s" % self.input_info)
        print_v("Input lists: %s" % self.input_lists)

        self.neuroML_arbor_recipe = NeuroML_Arbor_Recipe(
            self.nl_network,
            self.pop_indices_vs_gids,
            self.pops_vs_components,
            self.proj_weights,
            self.proj_delays,
            self.input_info,
            self.input_lists,
        )




# Create a NeuroML recipe
class NeuroML_Arbor_Recipe(arbor.recipe):
    def __init__(
        self,
        nl_network,
        pop_indices_vs_gids,
        pops_vs_components,
        proj_weights,
        proj_delays,
        input_info,
        input_lists,
    ):
        # The base C++ class constructor must be called first, to ensure that
        # all memory in the C++ class is initialized correctly.
        arbor.recipe.__init__(self)
        self.props = arbor.neuron_cable_properties()
        # self.cat = arbor.default_catalogue()
        # self.props.register(self.cat)
        self.pop_indices_vs_gids = pop_indices_vs_gids
        self.pops_vs_components = pops_vs_components
        self.nl_network = nl_network
        self.proj_weights = proj_weights
        self.proj_delays = proj_delays
        self.input_info = input_info
        self.input_lists = input_lists


    def create_arbor_cell(self, cell, gid, pop_id, index):

        if cell.arbor_cell == "cable_cell":

            default_tree = arbor.segment_tree()
            radius = (
                evaluate(cell.parameters["radius"], self.nl_network.parameters)
                if "radius" in cell.parameters
                else 3
            )

            default_tree.append(
                arbor.mnpos,
                arbor.mpoint(-1 * radius, 0, 0, radius),
                arbor.mpoint(radius, 0, 0, radius),
                tag=1,
            )

            labels = arbor.label_dict({"soma": "(tag 1)", "center": "(location 0 0.5)"})

            labels["root"] = "(root)"

            decor = arbor.decor()

            v_init = (
                evaluate(cell.parameters["v_init"], self.nl_network.parameters)
                if "v_init" in cell.parameters
                else -70
            )
            decor.set_property(Vm=v_init)

            decor.paint('"soma"', arbor.density(cell.parameters["mechanism"]))

            decor.place('"center"', arbor.spike_detector(0), "detector")

            for ip in self.input_info:
                if self.input_info[ip][0] == pop_id:
                    print_v("Stim: %s (%s) being placed on %s" % (ip,self.input_info[ip],pop_id))
                    for il in self.input_lists[ip]:
                        cellId, segId, fract, weight = il
                        if cellId == index:
                            if self.input_info[ip][1] == 'i_clamp': # TODO: remove hardcoding of this...
                                ic = arbor.iclamp(
                                    self.nl_network.parameters["input_del"],
                                    self.nl_network.parameters["input_dur"],
                                    self.nl_network.parameters["input_amp"],
                                )
                                print_v("Stim: %s on %s" % (ic,gid))
                                decor.place('"center"', ic, "iclamp")

            # (2) Mark location for synapse at the midpoint of branch 1 (the first dendrite).
            labels["synapse_site"] = "(location 0 0.5)"
            # (4) Attach a single synapse.
            decor.place('"synapse_site"', arbor.synapse("expsyn"), "syn")

            default_cell = arbor.cable_cell(default_tree, labels, decor)

            print_v("Created a new cell for gid %i: %s" % (gid, cell))
            print_v("%s" % (default_cell))

            return default_cell

    def get_pop_index(self, gid):
        # Todo: optimise...
        for pop_id in self.pop_indices_vs_gids:
            for index in self.pop_indices_vs_gids[pop_id]:
                if self.pop_indices_vs_gids[pop_id][index] == gid:
                    return pop_id, index

    def get_gid(self, pop_id, index):
        return self.pop_indices_vs_gids[pop_id][index]

    # (6) The num_cells method that returns the total number of cells in the model
    # must be implemented.
    def num_cells(self):
        ncells = sum(
            [
                len(self.pop_indices_vs_gids[pop_id])
                for pop_id in self.pop_indices_vs_gids
            ]
        )
        print_v("Getting num cells: %s" % (ncells))
        return ncells

    # (7) The cell_description method returns a cell
    def cell_description(self, gid):
        pop_id, index = self.get_pop_index(gid)
        comp = self.pops_vs_components[pop_id]

        a_cell = self.create_arbor_cell(
            self.nl_network.get_child(comp, "cells"), gid, pop_id, index
        )
        return a_cell


    # The kind method returns the type of cell with gid.
    # Note: this must agree with the type returned by cell_description.
    def cell_kind(self, gid):
        cell_kind = arbor.cell_kind.cable
        print_v("Getting cell_kind: %s" % (cell_kind))
        return cell_kind

    # (8) Make a ring network. For each gid, provide a list of incoming connections.
    def connections_on(self, gid):
        # Todo: optimise!!
        pop_id, index = self.get_pop_index(gid)
        conns = []
        for proj in self.nl_network.projections:
            if pop_id == proj.postsynaptic:
                w = self.proj_weights[proj.id]
                in_w = w.T[index]
                d = self.proj_delays[proj.id]
                in_d = d.T[index]
                print_v(
                    "Incoming connections for gid %i (%s[%s]), w: %s; d: %s"
                    % (gid, pop_id, index, in_w, in_d)
                )
                for src_index in range(len(in_w)):
                    if in_w[src_index] > 0:
                        src_gid = self.get_gid(proj.presynaptic, src_index)
                        conns.append(
                            arbor.connection(
                                (src_gid, "detector"),
                                "syn",
                                in_w[src_index],
                                in_d[src_index],
                            )
                        )

        print_v(
            "Making connections for gid %i (%s[%s]): %s" % (gid, pop_id, index, conns)
        )
        return conns

    def num_targets(self, gid):
        tot_in = len(self.connections_on(gid))
        print_v("num_targets for %i: %s" % (gid, tot_in))

        return 1

    def num_sources(self, gid):
        # Todo: optimise!!
        pop_id, index = self.get_pop_index(gid)
        tot_out = 0
        for proj in self.nl_network.projections:
            if pop_id == proj.presynaptic:
                w = self.proj_weights[proj.id]
                out_w = w[index]
                print_v(
                    "Outgoing connections for gid %i (%s[%s]), w: %s"
                    % (gid, pop_id, index, out_w)
                )
                for c in out_w:
                    if c > 0:
                        tot_out += 1

        print_v("num_sources for %i: %s" % (gid, tot_out))
        return 1

    # (9) Attach a generator to the first cell in the ring.
    def event_generators(self, gid):
        egs = []
        '''
        if gid == 0:
            sched = arbor.explicit_schedule([1])
            egs = [arbor.event_generator("syn", 0.1, sched)]'''

        print_v("Getting event_generators for %s: %s" % (gid, egs))
        return egs

    # (10) Place a probe at the root of each cell.
    def probes(self, gid):
        return [arbor.cable_probe_membrane_voltage('"root"')]

    def global_properties(self, kind):
        return self.props



from neuromllite import (
    Network,
    Cell,
    InputSource,
    Population,
    Synapse,
    RectangularRegion,
    RandomLayout,
)
from neuromllite import Projection, RandomConnectivity, Input, Simulation
import sys

################################################################################
###   Build new network

net = Network(id="Example4_PyNN")
net.notes = "Example 4: a network with PyNN cells & inputs"
net.parameters = {"input_amp": 0.99}

cell = Cell(id="testcell", pynn_cell="IF_cond_alpha")
cell.parameters = {"tau_refrac": 5, "i_offset": 0.1}
net.cells.append(cell)

cell2 = Cell(id="testcell2", pynn_cell="IF_cond_alpha")
cell2.parameters = {"tau_refrac": 5, "i_offset": -0.1}
net.cells.append(cell2)

input_source = InputSource(
    id="i_clamp",
    pynn_input="DCSource",
    parameters={"amplitude": "input_amp", "start": 200.0, "stop": 800.0},
)
net.input_sources.append(input_source)

r1 = RectangularRegion(id="region1", x=0, y=0, z=0, width=1000, height=100, depth=1000)
net.regions.append(r1)

p0 = Population(
    id="pop0",
    size=2,
    component=cell.id,
    properties={"color": "1 0 0", "radius": 20},
    random_layout=RandomLayout(region=r1.id),
)
p1 = Population(
    id="pop1",
    size=2,
    component=cell2.id,
    properties={"color": "0 1 0", "radius": 20},
    random_layout=RandomLayout(region=r1.id),
)
p2 = Population(
    id="pop2",
    size=1,
    component=cell2.id,
    properties={"color": "0 0 1", "radius": 20},
    random_layout=RandomLayout(region=r1.id),
)

net.populations.append(p0)
net.populations.append(p1)
net.populations.append(p2)

net.synapses.append(
    Synapse(
        id="ampaSyn",
        pynn_receptor_type="excitatory",
        pynn_synapse_type="cond_alpha",
        parameters={"e_rev": -10, "tau_syn": 2},
    )
)
net.synapses.append(
    Synapse(
        id="gabaSyn",
        pynn_receptor_type="inhibitory",
        pynn_synapse_type="cond_alpha",
        parameters={"e_rev": -80, "tau_syn": 10},
    )
)

net.projections.append(
    Projection(
        id="proj0",
        presynaptic=p0.id,
        postsynaptic=p1.id,
        synapse="ampaSyn",
        delay=2,
        weight=0.02,
    )
)
net.projections[0].random_connectivity = RandomConnectivity(probability=1)

net.projections.append(
    Projection(
        id="proj1",
        presynaptic=p0.id,
        postsynaptic=p2.id,
        synapse="gabaSyn",
        delay=2,
        weight=0.01,
    )
)
net.projections[1].random_connectivity = RandomConnectivity(probability=1)

net.inputs.append(
    Input(id="stim", input_source=input_source.id, population=p0.id, percentage=50)
)

print(net.to_json())
net_json_file = net.to_json_file("%s.json" % net.id)
net_yaml_file = net.to_yaml_file("%s.yaml" % net.id)


################################################################################
###   Build Simulation object & save as JSON

sim = Simulation(
    id="SimExample4",
    network=net_json_file,
    duration="1000",
    dt="0.01",
    record_traces={"all": "*"},
    record_spikes={"pop0": "*"},
)

sim.to_json_file()
sim.network = net_yaml_file
sim.to_yaml_file('%s.yaml'%sim.id)

sim.network = net_json_file  # reverting, for call below...


################################################################################
###   Run in some simulators

from neuromllite.NetworkGenerator import check_to_generate_or_run
import sys

check_to_generate_or_run(sys.argv, sim)



from neuromllite.utils import print_v
from neuromllite.DefaultNetworkHandler import DefaultNetworkHandler

from neuron import hoc

################################################################################
#
#    Not a full implementation of a native Neuron Handler yet!!
#
#    This will probably not evolve into a full Neuron handler, it is useful
#    though as a means of loading in and positioning cells for checking 3D
#    layout
#
################################################################################


class NeuronHandler(DefaultNetworkHandler):

    h = hoc.HocObject()

    globalPreSynId = 10000000

    preSectionsVsGids = dict()

    #
    #   Helper function for printing hoc before executing it
    #
    def executeHoc(self, command):

        cmdPrefix = "hoc >>>>>>>>>>: "

        if len(command) > 0:
            print_v(cmdPrefix + command)
            self.h(command)

    #
    #  Overridden from NetworkHandler
    #
    def handlePopulation(self, population_id, component, size=-1, component_obj=None):

        if size >= 0:
            sizeInfo = ", size " + str(size) + " cells"

            print_v(
                "Creating population: "
                + population_id
                + ", component: "
                + component
                + sizeInfo
            )

            self.executeHoc("{ n_" + population_id + " = " + str(size) + " }")
            self.executeHoc("{ n_" + population_id + "_local = 0 } ")
            self.executeHoc(
                "objectvar a_" + population_id + "[n_" + population_id + "]"
            )

        else:

            print_v(
                "Population: "
                + population_id
                + ", component: "
                + component
                + " specifies no size. Will lead to errors!"
            )

    #
    #  Overridden from NetworkHandler
    #
    def handleLocation(self, id, population_id, component, x, y, z):
        self.printLocationInformation(id, population_id, component, x, y, z)

        newCellName = population_id + "_" + str(id)

        createCall = (
            "new "
            + component
            + '("'
            + newCellName
            + '", "'
            + component
            + '", "New Cell: '
            + newCellName
            + " of type: "
            + component
            + '")'
        )

        cellInArray = "a_" + population_id + "[" + str(id) + "]"

        setupCreate = (
            "obfunc newCell() { {"
            + cellInArray
            + " = "
            + createCall
            + "} return "
            + cellInArray
            + " }"
        )

        self.executeHoc(setupCreate)

        newCell = self.h.newCell()

        newCell.position(float(x), float(y), float(z))

        self.executeHoc(
            "{n_" + population_id + "_local = n_" + population_id + "_local + 1}"
        )

        print_v(
            "Have just created cell: "
            + component
            + " at ("
            + str(x)
            + ", "
            + str(y)
            + ", "
            + str(z)
            + ")"
        )

    #
    #  Should be overridden to create cell group/population array
    #
    def handleProjection(
        self,
        projName,
        prePop,
        postPop,
        synapse,
        hasWeights=False,
        hasDelays=False,
        type="projection",
        synapse_obj=None,
        pre_synapse_obj=None,
    ):

        print_v("Projection: " + projName + " from " + prePop + " to " + postPop)

    #
    #  Overridden from NetworkHandler
    #
    def handleConnection(
        self,
        projName,
        id,
        prePop,
        postPop,
        synapseType,
        preCellId,
        postCellId,
        preSegId=0,
        preFract=0.5,
        postSegId=0,
        postFract=0.5,
        delay=0,
        weight=1,
    ):

        # self.printConnectionInformation(projName, id, prePop, postPop, synapseType, preCellId, postCellId, localWeight)

        print_v(
            "\n           --------------------------------------------------------------"
        )
        print_v(
            "Going to create a connection of type "
            + projName
            + ", id: "
            + str(id)
            + ", synapse type: "
            + synapseType
        )
        print_v(
            "From: "
            + prePop
            + ", id: "
            + str(preCellId)
            + ", segment: "
            + str(preSegId)
            + ", fraction: "
            + str(preFract)
        )
        print_v(
            "To  : "
            + postPop
            + ", id: "
            + str(postCellId)
            + ", segment: "
            + str(postSegId)
            + ", fraction: "
            + str(postFract)
        )
        print_v("   **** Connectivity not yet implemented!!  **** ")


##
##        threshold = 0
##
##        postPopCell = "a_"+postPop+"["+str(postCellId)+"]"
##        prePopCell = "a_"+prePop+"["+str(preCellId)+"]"
##
##
##
##        # Create post syn object
##
##        '''if self.h.isCellOnNode(str(postPop), int(postCellId)) == 1:'''
##
##        print_v("++++++++++++ PostCell: "+postPopCell+" is on this host...")
##
##        synObjName = projName+"_"+synapseType+"["+str(id)+"]"
##
##        ##########self.executeHoc("objref "+synObjName)
##        '''
#        self.executeHoc("{ "+postPopCell+".accessSectionForSegId("+str(postSegId)+") }")
#
#        self.executeHoc("{ "+"fractSecPost = "+postPopCell+".getFractAlongSection("+str(postFract)+", "+str(postSegId)+") }")
#
#        print_v("Synapse object at: "+str(h.fractSecPost) +" on sec: "+h.secname()+", or: "+str(postFract)+" on seg id: "+ str(postSegId))
#        #'''
##        self.executeHoc("{ "+synObjName+" = new "+synapseType+"(0) }")
##
##        self.executeHoc("{ "+postPopCell+".synlist.append("+synObjName+") }")
##
##        self.executeHoc("{ "+"localSynapseId = "+postPopCell+".synlist.count()-1 }")
##
##
##
##        # Create pre syn object
##
##
##        self.executeHoc("{ "+prePopCell+".accessSectionForSegId("+str(preSegId)+") }")
##        self.executeHoc("{ fractSecPre = "+prePopCell+".getFractAlongSection("+str(preFract)+", "+str(preSegId)+") }")
##
##        print_v("NetCon object at: "+str(h.fractSecPre) +" on sec: "+h.secname()+", or: "+str(preFract)+" on seg id: "+ str(preSegId))
##
##        self.executeHoc("{"+prePopCell+".synlist.append(new NetCon(&v(fractSecPre), " \
##                  +synObjName+", "+str(threshold)+", "+str(delay)+", "+str(weight)+")) }")
##
##        '''
#        else:
#
#            netConRef = "NetCon_"+str(self.globalPreSynId)
#            netConRefTemp = netConRef+"_temp"
#
#
#            preCellSegRef = str(prePopCell+"_"+str(preSegId))
#
#            gidToUse = self.globalPreSynId
#
#            if  preCellSegRef in self.preSectionsVsGids:
#                gidToUse = self.preSectionsVsGids[preCellSegRef]
#                print_v("Using *existing* NetCon with gid for pre syn: "+str(gidToUse)+"")
#            else:
#                print_v("Using new gid for pre syn: "+str(gidToUse)+"")
#                self.preSectionsVsGids[preCellSegRef] = self.globalPreSynId
#
#
#            if self.h.isCellOnNode(str(prePop), int(preCellId)) == 1:
#                print_v("++++++++++++ PreCell: "+prePopCell+" is here!!")
#
#                self.executeHoc("objref "+netConRef)
#                if  gidToUse == self.globalPreSynId:  # First time use of gid so create NetCon
#
#                    self.executeHoc("{ "+prePopCell+".accessSectionForSegId("+str(preSegId)+")"+" }")
#
#                    self.executeHoc("{ "+"pnm.pc.set_gid2node("+str(gidToUse)+", hostid)"+" }")
#                    self.executeHoc("{ "+netConRef+" = new NetCon(&v("+str(preFract) +"), nil)"+" }")
#
#                    self.executeHoc("{ "+netConRef+".delay = "+str(delayTotal)+" }")
#                    self.executeHoc("{ "+netConRef+".weight = "+str(localWeight)+" }")
#                    self.executeHoc("{ "+netConRef+".threshold = "+str(localThreshold)+" }")
#
#                    self.executeHoc("{ "+"pnm.pc.cell("+str(gidToUse)+", "+netConRef+")"+" }")
#
#
#
#            else:
#                print_v("------------ PreCell: "+prePopCell+" not on this host...")
#
#
#            # Connect pre to post
#
#
#            if self.isParallel == 1 and self.h.isCellOnNode(str(postPop), int(postCellId)) == 1:
#                self.executeHoc("objref "+netConRefTemp)
#                self.executeHoc(netConRefTemp+" = pnm.pc.gid_connect("+str(gidToUse)+","+postPopCell+".synlist.object(localSynapseId))")
#
#                self.executeHoc("{ "+netConRefTemp+".delay = "+str(delayTotal)+" }")
#                self.executeHoc("{ "+netConRefTemp+".weight = "+str(localWeight)+" }")
#                self.executeHoc("{ "+netConRefTemp+".threshold = "+str(localThreshold)+" }")
#
#
#            #self.executeHoc("netConInfoParallel("+netConRef+")")
#            #self.executeHoc("netConInfoParallel("+netConRefTemp+")")    #'''
##
##        self.globalPreSynId+=1



from neuromllite import (
    RandomLayout,
    Cell,
    Synapse,
    InputSource,
    Input,
    RectangularRegion,
)
from neuromllite.NetworkGenerator import generate_network
from neuromllite.utils import load_network_json
from neuromllite.DefaultNetworkHandler import DefaultNetworkHandler

################################################################################
###   Reuse network from Example1

net = load_network_json("Example1_TestNetwork.json")
net.id = "Example2_TestNetwork"

net.notes = (
    "A simple network with 2 populations & projection between them. "
    + "Cells are specified to be NeuroML 2 HH cell models & pre population "
    "is given a spiking input."
)


################################################################################
###   Add some elements to the network & save new JSON

r1 = RectangularRegion(id="region1", x=0, y=0, z=0, width=1000, height=100, depth=1000)
r2 = RectangularRegion(
    id="region2", x=0, y=200, z=0, width=1000, height=100, depth=1000
)
net.regions.append(r1)
net.regions.append(r2)

net.populations[0].random_layout = RandomLayout(region=r1.id)
net.populations[1].random_layout = RandomLayout(region=r2.id)

net.populations[0].component = "hhcell"
net.populations[1].component = "hhcell"

net.cells.append(Cell(id="hhcell", neuroml2_source_file="test_files/hhcell.cell.nml"))
net.synapses.append(
    Synapse(id="ampa", neuroml2_source_file="test_files/ampa.synapse.nml")
)

input_source = InputSource(
    id="poissonFiringSyn", neuroml2_source_file="test_files/inputs.nml"
)
net.input_sources.append(input_source)

net.inputs.append(
    Input(
        id="stim_%s" % net.populations[0].id,
        input_source=input_source.id,
        population=net.populations[0].id,
        percentage=80,
    )
)

print(net.to_json())
new_file = net.to_json_file("%s.json" % net.id)


################################################################################
###   Use a handler which just prints info on positions, etc.

def_handler = DefaultNetworkHandler()

generate_network(net, def_handler)


################################################################################
###   Export to some formats, e.g. try:
###        python Example2.py -graph2

from neuromllite.NetworkGenerator import check_to_generate_or_run
from neuromllite import Simulation
import sys

check_to_generate_or_run(sys.argv, Simulation(id="SimExample2", network=new_file))



from neuromllite import *
import sys
import json


from modelspec.base_types import EvaluableExpression

from modelspec.utils import (
    load_json,
    load_yaml,
    evaluate,
    parse_list_like,
    _parse_element,
    ascii_encode_dict,
)


verbose = False


def print_(text, print_it=False):
    """
    Print a message preceded by neuromllite, only if print_it=True
    """
    prefix = "neuromllite >>> "
    if not isinstance(text, str):
        text = ("%s" % text).decode("ascii")
    if print_it:
        print("%s%s" % (prefix, text.replace("\n", "\n" + prefix)))


def print_v(text):
    """
    Print a message preceded by neuromllite always
    """
    print_(text, True)


def load_network(filename):
    """
    Load a NeuroMLlite network JSON/YAML file
    """
    if filename.endswith(".yaml") or filename.endswith(".yml"):
        return load_network_yaml(filename)
    else:
        return load_network_json(filename)


def load_network_json(filename):
    """
    Load a NeuroMLlite network JSON file
    """

    data = load_json(filename)

    print_v("Loaded network specification from %s" % filename)

    net = Network.from_dict(data)

    # net = Network(id="net")
    # net = _parse_element(data, net)

    return net


def load_network_yaml(filename):
    """
    Load a NeuroMLlite network YAML file
    """

    data = load_yaml(filename)

    print_v("Loaded network specification from %s" % filename)

    net = Network.from_dict(data)

    # net = Network()
    # net = _parse_element(data, net)

    return net


def load_simulation_json(filename):
    """
    Load a NeuroMLlite simulation JSON file
    """


    with open(filename, "r") as f:
        data = json.load(f, object_hook=ascii_encode_dict)

    print_v("Loaded simulation specification from %s" % filename)

    sim = Simulation.from_dict(data)
    # sim = Simulation(id="sim")
    # sim = _parse_element(data, sim)

    return sim


def get_pops_vs_cell_indices_seg_ids(recordSpec, network):

    pvc = {}
    if recordSpec is not None:
        for p in recordSpec:
            indices = recordSpec[p]
            if p == "all":
                for pop in network.populations:
                    cell_indices_seg_ids = _generate_cell_indices_seg_ids(
                        pop.id, indices, network
                    )
                    pvc[pop.id] = cell_indices_seg_ids

            else:
                # pop = network.get_child(p, 'populations')
                cell_indices_seg_ids = _generate_cell_indices_seg_ids(
                    p, indices, network
                )
                pvc[p] = cell_indices_seg_ids

    return pvc


def _generate_cell_indices_seg_ids(pop_id, indices_segids, network):

    a = {}
    pop = network.get_child(pop_id, "populations")

    if not isinstance(indices_segids, str) or not ":" in indices_segids:
        seg_ids = None
        indices = indices_segids
    else:
        indices = indices_segids.split(":")[0]
        seg_id_info = indices_segids.split(":")[1]
        l = parse_list_like(seg_id_info)
        print_v("Parsed %s as %s" % (seg_id_info, l))
        seg_ids = l

    if indices == "*":
        size = evaluate(pop.size, network.parameters)
        for index in range(size):
            a[index] = seg_ids
    else:
        l = parse_list_like(indices)
        print_v("Parsed %s (full: %s) as %s" % (indices, indices_segids, l))
        for index in l:
            a[index] = seg_ids
    return a


def is_spiking_input_population(population, network):

    cell = network.get_child(population.component, "cells")

    return is_spiking_input_cell(cell)


def is_spiking_input_cell(cell):
    if cell.pynn_cell:
        if cell.pynn_cell == "SpikeSourcePoisson":
            return True
        else:
            return False


def is_spiking_input_nml_cell(component_obj):
    if component_obj.__class__.__name__ == "SpikeSourcePoisson":
        return True
    else:
        return False


def create_new_model(
    reference,
    duration,
    dt=0.025,  # ms
    temperature=6.3,  # degC
    default_region=None,
    parameters=None,
    cell_for_default_population=None,
    color_for_default_population="0.8 0 0",
    input_for_default_population=None,
    synapses=[],
    simulation_seed=12345,
    network_filename=None,
    simulation_filename=None,
):

    ################################################################################
    ###   Build a new network

    net = Network(id=reference)
    net.notes = "A network model: %s" % reference
    net.temperature = temperature  # degC
    if parameters:
        net.parameters = parameters

    ################################################################################
    ###   Add some regions

    if default_region:
        if type(default_region) == str:
            r1 = RectangularRegion(
                id=default_region, x=0, y=0, z=0, width=1000, height=100, depth=1000
            )
            net.regions.append(r1)
            default_region = r1
        else:
            net.regions.append(default_region)

    ################################################################################
    ###   Add some cells

    if cell_for_default_population:
        net.cells.append(cell_for_default_population)

    ################################################################################
    ###   Add some synapses

    for s in synapses:
        net.synapses.append(s)

    ################################################################################
    ###   Add some populations

    if cell_for_default_population:
        pop = Population(
            id="pop_%s" % cell_for_default_population.id,
            size=1,
            component=cell_for_default_population.id,
            properties={"color": color_for_default_population},
        )

        if default_region:
            pop.region = default_region

            pop.random_layout = RandomLayout(region=default_region.id)

        net.populations.append(pop)

    ################################################################################
    ###   Add a projection

    """
    net.projections.append(Projection(id='proj0',
                                      presynaptic=p0.id,
                                      postsynaptic=p1.id,
                                      synapse='ampa'))

    net.projections[0].random_connectivity=RandomConnectivity(probability=0.5)"""

    ################################################################################
    ###   Add some inputs

    if input_for_default_population:

        net.input_sources.append(input_for_default_population)

        net.inputs.append(
            Input(
                id="Stim_%s" % input_for_default_population.id,
                input_source=input_for_default_population.id,
                population=pop.id,
                percentage=100,
            )
        )

    ################################################################################
    ###   Save to JSON format

    net.id = reference

    print(net.to_json())
    if network_filename == None:
        network_filename = "%s.json" % net.id
    new_file = net.to_json_file(network_filename)

    ################################################################################
    ###   Build Simulation object & save as JSON

    sim = Simulation(
        id="Sim_%s" % reference,
        network=new_file,
        duration=duration,
        dt=dt,
        seed=simulation_seed,
        record_traces={"all": "*"},
    )

    if simulation_filename == None:
        simulation_filename = "%s.json" % sim.id
    sim.to_json_file(simulation_filename)

    return sim, net



from neuromllite import (
    Network,
    Cell,
    Population,
    Synapse,
    RectangularRegion,
    RandomLayout,
)
from neuromllite import Projection, RandomConnectivity, OneToOneConnector, Simulation

from neuromllite.NetworkGenerator import check_to_generate_or_run

import sys


def generate():

    dt = 0.025
    simtime = 1000

    ################################################################################
    ###   Build new network

    net = Network(id="Example7_Brunel2000")
    net.notes = "Example 7: based on network of Brunel 2000"

    net.parameters = {
        "g": 4,
        "eta": 1,
        "order": 5,
        "epsilon": 0.1,
        "J": 0.1,
        "delay": 1.5,
        "tauMem": 20.0,
        "tauSyn": 0.1,
        "tauRef": 2.0,
        "U0": 0.0,
        "theta": 20.0,
    }

    cell = Cell(id="ifcell", pynn_cell="IF_curr_alpha")

    cell.parameters = {
        "tau_m": "tauMem",
        "tau_refrac": "tauRef",
        "v_rest": "U0",
        "v_reset": "U0",
        "v_thresh": "theta",
        "cm": 0.001,
        "i_offset": 0,
    }

    # cell = Cell(id='hhcell', neuroml2_source_file='test_files/hhcell.cell.nml')
    net.cells.append(cell)

    poisson_input = Cell(id="poisson_input", pynn_cell="SpikeSourcePoisson")
    poisson_input.parameters = {
        "rate": "1000 * (eta*theta/(J*4*order*epsilon*tauMem)) * (4*order*epsilon)",
        "start": 0,
        "duration": 1e9,
    }
    net.cells.append(poisson_input)

    r1 = RectangularRegion(
        id="region1", x=0, y=0, z=0, width=1000, height=100, depth=1000
    )
    net.regions.append(r1)

    pE = Population(
        id="Epop",
        size="4*order",
        component=cell.id,
        properties={"color": ".9 0 0", "radius": 5},
        random_layout=RandomLayout(region=r1.id),
    )
    pEpoisson = Population(
        id="expoisson",
        size="4*order",
        component=poisson_input.id,
        properties={"color": "0.9 0.7 0.7", "radius": 3},
        random_layout=RandomLayout(region=r1.id),
    )
    pI = Population(
        id="Ipop",
        size="1*order",
        component=cell.id,
        properties={"color": "0 0 .9", "radius": 5},
        random_layout=RandomLayout(region=r1.id),
    )
    pIpoisson = Population(
        id="inpoisson",
        size="1*order",
        component=poisson_input.id,
        properties={"color": "0.7 0.7 0.9", "radius": 3},
        random_layout=RandomLayout(region=r1.id),
    )

    net.populations.append(pE)
    net.populations.append(pEpoisson)
    net.populations.append(pI)
    net.populations.append(pIpoisson)

    net.synapses.append(
        Synapse(
            id="ampa",
            pynn_receptor_type="excitatory",
            pynn_synapse_type="curr_alpha",
            parameters={"tau_syn": 0.1},
        )
    )

    net.synapses.append(
        Synapse(
            id="gaba",
            pynn_receptor_type="inhibitory",
            pynn_synapse_type="curr_alpha",
            parameters={"tau_syn": 0.1},
        )
    )

    delay_ext = dt

    downscale = 1
    J_eff = "J*%s" % (downscale)
    # synaptic weights, scaled for alpha functions, such that
    # for constant membrane potential, charge J would be deposited
    fudge = 0.00041363506632638  # ensures dV = J at V=0
    JE = "((%s)/tauSyn)*%s" % (J_eff, fudge)
    JI = "-1*g*%s" % (JE)

    net.projections.append(
        Projection(
            id="projEinput",
            presynaptic=pEpoisson.id,
            postsynaptic=pE.id,
            synapse="ampa",
            delay=delay_ext,
            weight=JE,
            one_to_one_connector=OneToOneConnector(),
        )
    )

    net.projections.append(
        Projection(
            id="projIinput",
            presynaptic=pIpoisson.id,
            postsynaptic=pI.id,
            synapse="ampa",
            delay=delay_ext,
            weight=JE,
            one_to_one_connector=OneToOneConnector(),
        )
    )

    net.projections.append(
        Projection(
            id="projEE",
            presynaptic=pE.id,
            postsynaptic=pE.id,
            synapse="ampa",
            delay="delay",
            weight=JE,
            random_connectivity=RandomConnectivity(probability="epsilon"),
        )
    )

    net.projections.append(
        Projection(
            id="projEI",
            presynaptic=pE.id,
            postsynaptic=pI.id,
            synapse="ampa",
            delay="delay",
            weight=JE,
            random_connectivity=RandomConnectivity(probability="epsilon"),
        )
    )

    net.projections.append(
        Projection(
            id="projIE",
            presynaptic=pI.id,
            postsynaptic=pE.id,
            synapse="gaba",
            delay="delay",
            weight=JI,
            random_connectivity=RandomConnectivity(probability="epsilon"),
        )
    )

    net.projections.append(
        Projection(
            id="projII",
            presynaptic=pI.id,
            postsynaptic=pI.id,
            synapse="gaba",
            delay="delay",
            weight=JI,
            random_connectivity=RandomConnectivity(probability="epsilon"),
        )
    )

    # print(net)
    # print(net.to_json())
    new_file = net.to_json_file("%s.json" % net.id)

    ################################################################################
    ###   Build Simulation object & save as JSON

    sim = Simulation(
        id="SimExample7",
        network=new_file,
        duration=simtime,
        dt=dt,
        seed=123,
        record_traces={pE.id: [0, 1], pI.id: [0, 1]},
        record_spikes={
            pE.id: "*",
            pI.id: "*",
            pEpoisson.id: [0, 1, 2, 3, 4],
            pIpoisson.id: [0, 1, 2, 3, 4],
        },
    )

    sim.to_json_file()

    return sim, net


if __name__ == "__main__":

    if "-sweep" in sys.argv:

        from neuromllite.sweep.ParameterSweep import *

        sim, net = generate()
        sim.record_traces = {}

        fixed = {"dt": 0.025, "order": 5}

        vary = {"eta": [0.5, 1, 1.5, 2, 3, 4, 5, 6, 7, 8, 9, 10]}
        # vary = {'eta':[1,2,5]}
        # vary['seed'] = [i for i in range(10)]
        vary["seed"] = [i for i in range(5)]

        simulator = "jNeuroML_NetPyNE"
        simulator = "jNeuroML_NEURON"
        simulator = "jNeuroML"
        simulator = "PyNN_NEST"
        simulator = "jNeuroML"

        nmllr = NeuroMLliteRunner("SimExample7.json", simulator=simulator)

        ps = ParameterSweep(
            nmllr,
            vary,
            fixed,
            num_parallel_runs=4,
            plot_all=False,
            heatmap_all=False,
            show_plot_already=False,
            peak_threshold=0,
        )

        report = ps.run()
        ps.print_report()

        #  ps.plotLines('weightInput','average_last_1percent',save_figure_to='average_last_1percent.png')
        # ps.plotLines('weightInput','mean_spike_frequency',save_figure_to='mean_spike_frequency.png')
        # ps.plotLines('eta','Einput[0]/spike:mean_spike_frequency',save_figure_to='mean_spike_frequency.png')
        ps.plotLines(
            "eta",
            "expoisson/0/poisson_input/spike:mean_spike_frequency",
            second_param="seed",
            save_figure_to="mean_spike_frequency_ein.png",
        )
        ps.plotLines(
            "eta",
            "inpoisson/0/poisson_input/spike:mean_spike_frequency",
            second_param="seed",
            save_figure_to="mean_spike_frequency_iin.png",
        )
        ps.plotLines(
            "eta",
            "Epop/0/ifcell/spike:mean_spike_frequency",
            second_param="seed",
            save_figure_to="mean_spike_frequency_e.png",
        )
        ps.plotLines(
            "eta",
            "Ipop/0/ifcell/spike:mean_spike_frequency",
            second_param="seed",
            save_figure_to="mean_spike_frequency_i.png",
        )

        import matplotlib.pyplot as plt

        plt.show()

    if "-sweep2" in sys.argv:

        sim, net = generate()
        sim.record_traces = {}

        fixed = {"dt": 0.025, "order": 5}

        vary = {"eta": [0.5, 1, 1.5, 2, 3]}
        vary = {"epsilon": [0.01, 0.1, 0.2, 0.5, 0.9]}
        vary = {"J": [0.01, 0.1, 0.2, 0.5, 0.9]}
        # vary = {'g':[0.5,1,1.5,2,3,20]}

        first = vary.keys()[0]

        # vary = {'eta':[1,2,5]}
        # vary['seed'] = [i for i in range(10)]
        vary["seed"] = [i for i in range(3)]

        simulator = "jNeuroML"
        simulator = "PyNN_NEST"
        simulator = "jNeuroML_NetPyNE"
        simulator = "jNeuroML"
        simulator = "jNeuroML_NEURON"

        nmllr = NeuroMLliteRunner("SimExample7.json", simulator=simulator)

        ps = ParameterSweep(
            nmllr,
            vary,
            fixed,
            num_parallel_runs=16,
            plot_all=False,
            heatmap_all=False,
            show_plot_already=False,
            peak_threshold=0,
        )

        report = ps.run()
        ps.print_report()

        #  ps.plotLines('weightInput','average_last_1percent',save_figure_to='average_last_1percent.png')
        # ps.plotLines('weightInput','mean_spike_frequency',save_figure_to='mean_spike_frequency.png')
        # ps.plotLines('eta','Einput[0]/spike:mean_spike_frequency',save_figure_to='mean_spike_frequency.png')

        second = "seed"
        ps.plotLines(
            first,
            "expoisson/0/poisson_input/spike:mean_spike_frequency",
            second_param=second,
            save_figure_to="mean_spike_frequency_ein.png",
        )
        ps.plotLines(
            first,
            "inpoisson/0/poisson_input/spike:mean_spike_frequency",
            second_param=second,
            save_figure_to="mean_spike_frequency_iin.png",
        )
        ps.plotLines(
            first,
            "Epop/0/ifcell/spike:mean_spike_frequency",
            second_param=second,
            save_figure_to="mean_spike_frequency_e.png",
        )
        ps.plotLines(
            first,
            "Ipop/0/ifcell/spike:mean_spike_frequency",
            second_param=second,
            save_figure_to="mean_spike_frequency_i.png",
        )

        import matplotlib.pyplot as plt

        plt.show()

    else:

        sim, net = generate()

        ################################################################################
        ###   Run in some simulators

        import sys

        check_to_generate_or_run(sys.argv, sim)



import os

from neuroml.hdf5.NetworkContainer import *
from neuromllite import NetworkReader

from neuromllite.utils import print_v, load_json


class PsyNeuLinkReader(NetworkReader):

    component_objects = {}  # Store cell ids vs objects, e.g. NeuroML2 based object

    PRE_SYN = "silentSyn"
    POST_SYN = "rsDL"

    def __init__(self, **parameters):

        print_v("Creating PsyNeuLinkReader with %s..." % parameters)
        self.parameters = parameters
        self.current_population = None
        self.pre_pop = None
        self.post_pop = None

    def _generate_id(self, name):
        return (
            name.replace(" ", "_").replace("[", "_").replace("]", "_").replace("-", "_")
        )

    def parse(self, handler):

        FORMAT = "BIDS-MDF"
        filename = os.path.abspath(self.parameters["filename"])
        id = filename.split("/")[-1].split(".")[0]

        self.handler = handler

        bids_mdf = load_json(filename)

        print(bids_mdf)

        if len(bids_mdf["graphs"]) != 1:
            raise Exception("Can only parse %s files with a single graph" % (FORMAT))

        graph = bids_mdf["graphs"][0]
        net_id = self._generate_id(graph["name"])

        if "id" in self.parameters:
            id = self._generate_id(self.parameters["id"])
        else:
            id = net_id

        notes = "Network (%s) read in from PsyNeuLink/%s: %s" % (id, FORMAT, filename)
        handler.handle_document_start(id, notes)

        handler.handle_network(id, notes)

        for node_id in graph["nodes"]:
            node = graph["nodes"][node_id]
            pop_id = self._generate_id(node["name"])
            size = 1
            properties = {}
            self.handler.handle_population(
                pop_id, self.parameters["DEFAULT_CELL_ID"], size, properties=properties
            )

        for edge_id in graph["edges"]:
            edge = graph["edges"][edge_id]
            pre_pop = self._generate_id(edge["sender"])
            post_pop = self._generate_id(edge["receiver"])
            weight = float(edge["weight"]) if "weight" in edge else 1
            proj_id = self._generate_id(edge_id)
            self.handler.handle_projection(
                proj_id,
                pre_pop,
                post_pop,
                synapse=self.PRE_SYN,
                type="continuousProjection",
            )

            self.handler.handle_connection(
                proj_id,
                0,
                pre_pop,
                post_pop,
                self.PRE_SYN,
                0,
                0,
                delay=0,
                weight=weight,
            )

    """
    def parse_group(self, g):
        #print("+++++++++++++++Parsing group: "+ str(g)+", name: "+g._v_name)

        for node in g:
            #print("   ------Sub node: %s, class: %s, name: %s (parent: %s)"   % (node,node._c_classid,node._v_name, g._v_name))

            if node._c_classid == 'GROUP':
                if g._v_name=='populations':
                    pop_id = node._v_name.replace('-','_')
                    self.current_population = pop_id
                    self.pop_locations[self.current_population] = {}

                if g._v_name=='connectivity':
                    self.pre_pop = node._v_name.replace('-','_')
                    self.post_pop=None
                    #print("Conn %s -> ?"%(self.pre_pop))
                elif self.pre_pop!=None and self.post_pop==None:
                    self.post_pop = node._v_name.replace('-','_')
                    #print("Conn2 %s -> %s"%(self.pre_pop,self.post_pop))
                elif self.pre_pop!=None and self.post_pop!=None:
                    #print("Conn3 %s -> %s"%(self.pre_pop,self.post_pop))
                    pass

                self.parse_group(node)

            if self._is_dataset(node):
                self.parse_dataset(node)

        self.current_population = None


    def _is_dataset(self, node):
          return node._c_classid == 'ARRAY' or node._c_classid == 'CARRAY'


    def _is_interneuron(self, pop_id):
        if 'PC' in pop_id or 'SS' in pop_id or 'SP' in pop_id:
            return False
        else:
            return True


    def parse_dataset(self, d):
        #print_v("Parsing dataset/array: "+ str(d))

        # Population
        if self.current_population and d.name=='locations':

            perc_cells = self.parameters['percentage_cells_per_pop'] if 'percentage_cells_per_pop' in self.parameters else 100
            if perc_cells>100: perc_cells = 100

            size = max(0,int((perc_cells/100.)*d.shape[0]))

            if size>0:
                properties = {}
                if self._is_interneuron(self.current_population):
                    properties['radius'] = 5
                    type='I'
                else:
                    properties['radius'] = 10
                    type='E'
                properties['type'] = type

                layer = self.current_population.split('_')[0]
                properties['region'] = layer
                try:
                    import opencortex.utils.color as occ
                    if layer == 'L23':
                        if type=='E': color = occ.L23_PRINCIPAL_CELL
                        if type=='I': color = occ.L23_INTERNEURON
                    if layer == 'L4':
                        if type=='E': color = occ.L4_PRINCIPAL_CELL
                        if type=='I': color = occ.L4_INTERNEURON
                    if layer == 'L5':
                        if type=='E': color = occ.L5_PRINCIPAL_CELL
                        if type=='I': color = occ.L5_INTERNEURON
                    if layer == 'L6':
                        if type=='E': color = occ.L6_PRINCIPAL_CELL
                        if type=='I': color = occ.L6_INTERNEURON

                    properties['color'] = color
                except:
                    # Don't worry about it, it's just metadata
                    pass

                component_obj = None

                if self.parameters['DEFAULT_CELL_ID'] in self.component_objects:
                    component_obj = self.component_objects[self.parameters['DEFAULT_CELL_ID']]
                else:
                    if 'cell_info' in self.parameters:
                        def_cell_info = self.parameters['cell_info'][self.parameters['DEFAULT_CELL_ID']]
                        if def_cell_info.neuroml2_source_file:
                            from pyneuroml import pynml
                            nml2_doc = pynml.read_neuroml2_file(def_cell_info.neuroml2_source_file,
                                                    include_includes=True)
                            component_obj = nml2_doc.get_by_id(self.parameters['DEFAULT_CELL_ID'])
                            print_v("Loaded NeuroML2 object %s from %s "%(component_obj,def_cell_info.neuroml2_source_file))
                            self.component_objects[self.parameters['DEFAULT_CELL_ID']] = component_obj


                self.handler.handle_population(self.current_population,
                                         self.parameters['DEFAULT_CELL_ID'],
                                         size,
                                         component_obj=component_obj,
                                         properties=properties)

                print_v("   There are %i cells in: %s"%(size, self.current_population))
                for i in range(0, d.shape[0]):

                    if i<size:
                        row = d[i,:]
                        x = row[0]
                        y = row[1]
                        z = row[2]
                        self.pop_locations[self.current_population][i]=(x,y,z)
                        self.handler.handle_location(i, self.current_population, self.parameters['DEFAULT_CELL_ID'], x, y, z)


        # Projection
        elif self.pre_pop!=None and self.post_pop!=None:

            proj_id = 'Proj__%s__%s'%(self.pre_pop,self.post_pop)
            synapse = 'gaba'
            if 'PC' in self.pre_pop or 'SS' in self.pre_pop: # TODO: better choice between E/I cells
                synapse = 'ampa'

            (ii, jj) = np.nonzero(d)
            conns_here = False
            pre_num = len(self.pop_locations[self.pre_pop]) if self.pre_pop in self.pop_locations else 0
            post_num = len(self.pop_locations[self.post_pop]) if self.post_pop in self.pop_locations else 0

            if pre_num>0 and post_num>0:
                for index in range(len(ii)):
                    if ii[index]<pre_num and \
                       jj[index]<post_num:
                        conns_here=True
                        break

                if conns_here:
                    print_v("Conn %s -> %s (%s)"%(self.pre_pop,self.post_pop, synapse))
                    self.handler.handle_projection(proj_id,
                                         self.pre_pop,
                                         self.post_pop,
                                         synapse)

                    conn_count = 0

                    for index in range(len(ii)):
                        i = ii[index]
                        j = jj[index]
                        if i<pre_num and j<post_num:
                            #print("  Conn5 %s[%s] -> %s[%s]"%(self.pre_pop,i,self.post_pop,j))
                            delay = 1.111
                            weight =1
                            self.handler.handle_connection(proj_id,
                                             conn_count,
                                             self.pre_pop,
                                             self.post_pop,
                                             synapse, \
                                             i, \
                                             j, \
                                             delay = delay, \
                                             weight = weight)
                            conn_count+=1


                    self.handler.finalise_projection(proj_id,
                                         self.pre_pop,
                                         self.post_pop,
                                         synapse)

            self.post_pop=None"""


if __name__ == "__main__":

    test_files = [
        "../../neuroConstruct/osb/showcase/PsyNeuLinkShowcase/PsyNeuLink/model_with_simple_graph.json"
    ]
    test_files.append(
        "../../neuroConstruct/osb/showcase/PsyNeuLinkShowcase/NeuroML2/ABC.bids-mdf.json"
    )
    test_files.append(
        "../../neuroConstruct/osb/showcase/PsyNeuLinkShowcase/PsyNeuLink/tests/ColorMotionTask_SIMPLE.json"
    )
    test_files.append(
        "../../neuroConstruct/osb/showcase/PsyNeuLinkShowcase/PsyNeuLink/tests/StroopModelEVC.json"
    )
    test_files.append(
        "../../neuroConstruct/osb/showcase/PsyNeuLinkShowcase/PsyNeuLink/tests/GilzenratModel.json"
    )
    test_files.append(
        "../../neuroConstruct/osb/showcase/PsyNeuLinkShowcase/NeuroML2/ABCD.bids-mdf.json"
    )

    test_files = [
        "../../neuroConstruct/osb/showcase/PsyNeuLinkShowcase/NeuroML2/ABCD.bids-mdf.json"
    ]
    test_files = [
        "../../neuroConstruct/osb/showcase/PsyNeuLinkShowcase/NeuroML2/FN.bids-mdf.json"
    ]

    for filename in test_files:
        pnl_parser = PsyNeuLinkReader(filename=filename, DEFAULT_CELL_ID="hhcell")

        from neuromllite.DefaultNetworkHandler import DefaultNetworkHandler

        def_handler = DefaultNetworkHandler()

        pnl_parser.parse(def_handler)

        from neuroml.hdf5.NetworkBuilder import NetworkBuilder

        neuroml_handler = NetworkBuilder()

        pnl_parser = PsyNeuLinkReader(filename=filename, DEFAULT_CELL_ID="hhcell")
        pnl_parser.parse(neuroml_handler)

        from neuroml.writers import NeuroMLWriter

        nml_file_name = "%s.net.nml" % neuroml_handler.get_nml_doc().id

        NeuroMLWriter.write(neuroml_handler.get_nml_doc(), nml_file_name)

        print("Written NeuroML file to: %s" % nml_file_name)
